#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "i8085.h"
#include "t8085.h"

#include "dis.h"

#define COLUMN_OPCODE	32
#define COLUMN_ARGUMENT 38
#define COLUMN_COMMENT	56

memopt_t memopt[0x10000];	/* memory flags and options */

sym_t	* symbols = NULL;
com_t	* comments = NULL;
dis_t	dis = {{0,}, };
I8085	i8085 = {{{0,},}, };

/***************************************************************************** */
/* print a value in 2 or 3 digit hexadecimal with trailing 'h' */
/*******************************************************************---------- */
char    * hexb(int val)
{
	static	int  hexb_indx=0;
	static	char hexb_buff[6*16];
	char	* dst = hexb_buff + 6 * (++hexb_indx & 15);

	if (val > 0x9F)
		sprintf(dst, "%03XH", val & 255);
	else
	if (val < 10)
		sprintf(dst, "%d", val);
	else
		sprintf(dst, "%02XH", val);
	return dst;
}

/***************************************************************************** */
/* print a value in 4 or 5 digit hexadecimal with trailing 'h' */
/*******************************************************************---------- */
char    * hexw(int val)
{
	static	int  hexw_indx=0;
	static	char hexw_buff[8*16];
	char	* dst = hexw_buff + 8 * (++hexw_indx & 15);
	if (val > 0x9FFF)
		sprintf(dst, "%05XH", val & 65535);
	else
		sprintf(dst, "%04XH", val);
	return dst;
}

/***************************************************************************** */
/* convert hexadecimal digit sequence into a u16 */
/*******************************************************************---------- */
u16 hextoi(char * src)
{
	u16	rc = 0;
	while (*src) {
		if ((*src >= '0') && (*src <= '9'))
			rc = rc * 16 + *src - '0';
		else
		if ((*src >= 'A') && (*src <= 'F'))
			rc = rc * 16 + *src - 'A' + 10;
		else
		if ((*src >= 'a') && (*src <= 'f'))
			rc = rc * 16 + *src - 'a' + 10;
		else
			return rc;
		src++;
	}
	return rc;
}

/***************************************************************************** */
/* define an option for a memory location */
/*******************************************************************---------- */
void    def_opt(u16 addr, unsigned mode, unsigned size, unsigned opt)
{
	memopt[addr].m.mode = mode;
	memopt[addr].m.size = size;
	memopt[addr].m.opt  = opt;
}


/***************************************************************************** */
/* add a symbol to our table */
/*******************************************************************---------- */
void	add_sym(u16 addr, symtype_t type, char * name)
{
	sym_t *sym, * sym0, * sym1;
	for (sym0 = NULL, sym1 = symbols; (sym1); sym0 = sym1, sym1 = sym1->next)
		if (addr <= sym1->addr)
			break;
	sym = (sym_t *) malloc(sizeof(sym_t) + strlen(name));

	if (!sym)
		return;

	sym->next = sym1;
	sym->type = type;
	sym->addr = addr;
	strcpy(sym->name, name);

	if (sym0)
		sym0->next = sym;
	else
		symbols    = sym;

	memopt[addr].m.symbol = 1;
}

/***************************************************************************** */
/* add a comment to our table */
/*******************************************************************---------- */
void	com_t_add(u16 addr, comtype_t type, char * text)
{
	com_t     * com, * com0, * com1;

	for (com0 = NULL, com1 = comments; (com1); com0 = com1, com1 = com1->next)
		if (addr <= com1->addr)
			break;

	if ((com1) &&		   /* found an entry */
	    (com1->addr == addr) &&     /* address is equal ? */
	    (com1->type == type) &&     /* address is equal ? */
	    strncmp(com1->text,text,strlen(text)))	  /* and text not yet found ? */
	{
		com = (com_t *) realloc(com1, sizeof(com_t) + strlen(com1->text) + 3 + strlen(text) + 1);

		if (!com)
			return;

		strcat(com->text, "\n; ");
		strcat(com->text, text);

		if (com0)
			com0->next = com;
		else
			comments   = com;
		return;
	}

	com = (com_t *) malloc(sizeof(com_t) + strlen(text));

	if (!com)
		return;

	com->next = com1;
	com->type = type;
	com->addr = addr;
	strcpy(com->text, text);

	if (com0)
		com0->next = com;
	else
		comments   = com;

	memopt[addr].m.comment = 1;
}

/***************************************************************************** */
/* get a symbol name for the given address and of the given type */
/*******************************************************************---------- */
char	* get_sym(u16 addr, symtype_t type)
{
	static	char sym_buff[128];
	sym_t  *sym;
	if (memopt[addr].m.symbol) {
		for (sym = symbols; (sym); sym = sym->next) {
			if (addr != sym->addr)
				continue;
			if ((type == SYM_NONE) || (type == sym->type))
				return sym->name;

		}
	}
	if (addr > 0 && memopt[addr-1].m.symbol) {
		for (sym = symbols; (sym); sym = sym->next) {
			if (addr != sym->addr+1)
				continue;
			if ((type == SYM_NONE) || (type == sym->type)) {
				sprintf(sym_buff, "%s+1", sym->name);
				return sym_buff;
			}

		}
		if (addr > 1 && memopt[addr-2].m.symbol) {
			for (sym = symbols; (sym); sym = sym->next) {
				if (addr != sym->addr+2)
					continue;
				if ((type == SYM_NONE) || (type == sym->type)) {
					sprintf(sym_buff, "%s+2", sym->name);
					return sym_buff;
				}

			}
			if (addr > 2 && memopt[addr-3].m.symbol) {
				for (sym = symbols; (sym); sym = sym->next) {
					if (addr != sym->addr+3)
						continue;
					if ((type == SYM_NONE) || (type == sym->type)) {
						sprintf(sym_buff, "%s+3", sym->name);
						return sym_buff;
					}

				}
			}
		}
	}
	return hexw(addr);
}

/***************************************************************************** */
/* get a comment for the given address */
/*******************************************************************---------- */
char *com_get(u16 addr, comtype_t type)
{
	com_t	*com;
	if (memopt[addr].m.comment) {
		for (com = comments; (com); com = com->next) {
			if (addr != com->addr)
				continue;
			if (type == COM_NONE || type == com->type)
				return com->text;
		}
	}
	return "";
}

/***************************************************************************** */
/* read a symbol, flags and options definitions file */
/*******************************************************************---------- */
void def_file(char *filename)
{
	FILE *filp;
	char *line;
	char buff[64];
	char *src;
	u16 addr;
	u16 offs;

	if (NULL == (line = alloca(1024)))
		return;

	filp = fopen(filename, "r");

	if (!filp)
		return;

	while (!feof(filp)) {
		if (NULL == fgets(line, 1024, filp))
			break;

		src = strtok(line, "\t \n\x1A");

		if (!src)
			continue;

		addr = hextoi(src);
		src = strtok(NULL, "\t \n\x1A");

		if (!src)
			continue;

		switch (*src) {
		case '+':
			src = strtok(NULL, "\n\x1A");
			if (src)
				com_t_add(addr, COM_APPEND, src);
			break;

		case ';':
			src = strtok(NULL, "\n\x1A");
			if (src)
				com_t_add(addr, COM_BLOCK, src);
			break;

		case '=':
			src = strtok(NULL, "\n\x1A");
			if (src)
				add_sym(addr, SYM_NONE, src);
			break;

		case 'c': case 'C':
			def_opt(addr, MEM_CODE, 0, 0);
			src = strtok(NULL, "\t \n\x1A");
			if (src)
				add_sym(addr, SYM_CODE, src);
			break;

		case 'b': case 'B':
			def_opt(addr, MEM_DATA, 0, OPT_BYTES);
			memopt[addr].m.opt  = OPT_BYTES;
			src = strtok(NULL, "\t \n\x1A");
			if (src)
				add_sym(addr, SYM_DATA, src);
			break;


		case '7':
			def_opt(addr, MEM_DATA, 0, OPT_7BIT);
			src = strtok(NULL, "\t \n\x1A");
			if (src)
				add_sym(addr, SYM_DATA, src);
			break;

		case 'a': case 'A':
			def_opt(addr, MEM_DATA, 0, OPT_ASCII);
			memopt[addr].m.opt  = OPT_ASCII;
			src = strtok(NULL, "\t \n\x1A");
			if (src)
				add_sym(addr, SYM_DATA, src);
			break;

		case 'w': case 'W':
			def_opt(addr, MEM_DATA, 1, OPT_WORDS);
			src = strtok(NULL, "\t \n\x1A");
			if (src)
				add_sym(addr, SYM_DATA, src);
			break;

		case 'e': case 'E':
			def_opt(addr, MEM_DATA, 1, OPT_ENTRY);
			src = strtok(NULL, "\t \n\x1A");
			if (src)
				add_sym(addr, SYM_DATA, src);
			break;

		case 's': case 'S':
			def_opt(addr, MEM_DATA, 2, OPT_FSNG);
			src = strtok(NULL, "\t \n\x1A");
			if (src)
				add_sym(addr, SYM_DATA, src);
			break;

		case 'd': case 'D':
			def_opt(addr, MEM_DATA, 3, OPT_FDBL);
			src = strtok(NULL, "\t \n\x1A");
			if (src)
				add_sym(addr, SYM_DATA, src);
			break;

		case 'o': case 'O':
			offs = mem[addr] + 256 * mem[addr+1];
			switch (src[1]) {
			case 'C':
				def_opt(addr, MEM_DATA, 1, OPT_OC);
				memopt[addr].m.opt  = OPT_OC;
				if (memopt[offs].m.mode == MEM_UNKNOWN) {
					def_opt(offs, MEM_CODE, 0, OPT_NONE);
					if (memopt[offs].m.comment == 0) {
						sprintf(buff, "CODE xref [%04Xh]", addr);
						com_t_add(offs, COM_APPEND, buff);
					}
				}
				break;
			case '7':
				def_opt(addr, MEM_DATA, 1, OPT_OD7);
				if (memopt[offs].m.mode == MEM_UNKNOWN) {
					def_opt(offs, MEM_DATA, 0, OPT_7BIT);
					if (memopt[offs].m.comment == 0) {
						sprintf(buff, "7BIT xref [%04Xh]", addr);
						com_t_add(offs, COM_APPEND, buff);
					}
				}
				break;
			case 'A':
				def_opt(addr, MEM_DATA, 1, OPT_ODA);
				if (memopt[offs].m.mode == MEM_UNKNOWN) {
					def_opt(offs, MEM_DATA, 0, OPT_ASCII);
					if (memopt[offs].m.comment == 0) {
						sprintf(buff, "ASCII xref [%04Xh]", addr);
						com_t_add(offs, COM_APPEND, buff);
					}
				}
				break;
			case 'B':
				def_opt(addr, MEM_DATA, 1, OPT_ODB);
				if (memopt[offs].m.mode == MEM_UNKNOWN) {
					def_opt(offs, MEM_DATA, 0, OPT_BYTES);
					if (memopt[offs].m.comment == 0) {
						sprintf(buff, "BYTES xref [%04Xh]", addr);
						com_t_add(offs, COM_APPEND, buff);
					}
				}
				break;
			case 'W':
				def_opt(addr, MEM_DATA, 1, OPT_ODW);
				if (memopt[offs].m.mode == MEM_UNKNOWN) {
					def_opt(offs, MEM_DATA, 1, OPT_WORDS);
					if (memopt[offs].m.comment == 0) {
						sprintf(buff, "WORDS xref [%04Xh]", addr);
						com_t_add(offs, COM_APPEND, buff);
					}
				}
				break;
			case 'E':
				def_opt(addr, MEM_DATA, 1, OPT_ODE);
				if (memopt[offs].m.mode == MEM_UNKNOWN) {
					def_opt(offs, MEM_DATA, 1, OPT_ENTRY);
					if (memopt[offs].m.comment == 0) {
						sprintf(buff, "ENTRY xref [%04Xh]", addr);
						com_t_add(offs, COM_APPEND, buff);
					}
				}
				break;
			default:
				def_opt(addr, MEM_DATA, 1, OPT_ODB);
				if (memopt[offs].m.mode == MEM_UNKNOWN) {
					def_opt(offs, MEM_DATA, 0, OPT_BYTES);
				}
				break;
			}
			src = strtok(NULL, "\t \n\x1A");
			if (src)
				add_sym(addr, SYM_DATA, src);
			break;

		}
	}
	fclose(filp);
}

/***************************************************************************** */
/* pad the output buffer with blanks and output an opcode */
/*******************************************************************---------- */
int     pad_opcode(char * fmt, ...)
{
	va_list arg;
	int	len;

	while (dis.dst < dis.buff + COLUMN_OPCODE)
		*dis.dst++ = ' ';
	va_start(arg, fmt);
	len = vsprintf(dis.dst, fmt, arg);
	va_end(arg);
	dis.dst += len;
	return len;
}

/***************************************************************************** */
/* pad the output buffer with blanks and output an argument */
/*******************************************************************---------- */
int     pad_argument(char * fmt, ...)
{
	va_list arg;
	int	len;

	while (dis.dst < dis.buff + COLUMN_ARGUMENT)
		*dis.dst++ = ' ';
	va_start(arg, fmt);
	len = vsprintf(dis.dst, fmt, arg);
	va_end(arg);
	dis.dst += len;
	return len;
}

/***************************************************************************** */
/* pad the output buffer with blanks and output a comment */
/*******************************************************************---------- */
int     pad_comment(char * fmt, ...)
{
	va_list arg;
	int	len;

	while (dis.dst < dis.buff + COLUMN_COMMENT)
		*dis.dst++ = ' ';

	if (!dis.padcom) {
		dis.padcom = 1;
		*dis.dst++ = ';';
	}

	va_start(arg, fmt);
	len = vsprintf(dis.dst, fmt, arg);
	va_end(arg);
	dis.dst += len;
	return len;
}

/***************************************************************************** */
/* disassemble an destination operand from the given table entry */
/*******************************************************************---------- */
int     dis_dst(TABLE * ptab, int comma)
{
	int	len = 0;

	switch (ptab->dst)
	{
	case _d0:
		return 0;
	case _dB:
		len = pad_argument("%s","B");
		break;
	case _dC:
		len = pad_argument("%s","C");
		break;
	case _dD:
		len = pad_argument("%s","D");
		break;
	case _dE:
		len = pad_argument("%s","E");
		break;
	case _dH:
		len = pad_argument("%s","H");
		break;
	case _dL:
		len = pad_argument("%s","L");
		break;
	case _dM:
		dis.flag.dst_mem8 = 1;
		dis.flag.dst_addr = 1;
		dis.addr = i8085.hl.r16;
		dis.mem8 = mem[dis.addr];
		len = pad_argument("%s","M");
		break;
	case _dA:
		len = pad_argument("%s","A");
		break;
	case _dIM:
		break;
	case _dMBC:
		dis.flag.dst_mem8 = 1;
		dis.flag.dst_addr = 1;
		dis.addr = i8085.bc.r16;
		dis.mem8 = mem[dis.addr];
		len = pad_argument("%s","B");
		break;
	case _dMDE:
		dis.flag.dst_mem8 = 1;
		dis.flag.dst_addr = 1;
		dis.addr = i8085.de.r16;
		dis.mem8 = mem[dis.addr];
		len = pad_argument("%s","D");
		break;
	case _dMSP:
		dis.flag.dst_mem16 = 1;
		dis.flag.dst_addr  = 1;
		dis.addr = i8085.sp.r16;
		dis.mem16 = mem[dis.addr] + 256 * mem[dis.addr+1];
		len = pad_argument("%s","(SP)");
		break;
	case _dIND8:
		if (!dis.flag.has_ind) {
			dis.flag.has_ind = 1;
			dis.ind16 = mem[dis.pc++];
			dis.ind16 += 256 * mem[dis.pc++];
		}
		dis.flag.dst_mem8 = 1;
		dis.addr = dis.ind16;
		dis.mem8 = mem[dis.addr];
		len = pad_argument("%s", get_sym(dis.ind16, SYM_NONE));
		break;
	case _dIMM8:
		if (!dis.flag.has_imm8) {
			dis.flag.has_imm8 = 1;
			dis.imm8  = mem[dis.pc++];
		}
		len = pad_argument("%s", hexb(dis.imm8));
		break;
	case _dPORT8:
		if (!dis.flag.has_port) {
			dis.flag.has_port = 1;
			dis.port  = mem[dis.pc++];
		}
		dis.flag.dst_port = 1;
		len = pad_argument("%s", hexb(dis.port));
		break;
	case _dBC:
		len = pad_argument("%s","B");
		break;
	case _dDE:
		len = pad_argument("%s","D");
		break;
	case _dHL:
		len = pad_argument("%s","H");
		break;
	case _dSP:
		len = pad_argument("%s","SP");
		break;
	case _dAF:
		len = pad_argument("%s","PSW");
		break;
	case _dPC:
		len = pad_argument("%s","$");
		break;
	case _dPCHL:
		dis.flag.dst_addr = 1;
		dis.addr  = i8085.hl.r16;
		break;
	case _dIND16:
		if (!dis.flag.has_ind) {
			dis.flag.has_ind = 1;
			dis.ind16 = mem[dis.pc++];
			dis.ind16 += 256 * mem[dis.pc++];
		}
		dis.flag.dst_mem16 = 1;
		dis.addr  = dis.ind16;
		dis.mem16 = mem[dis.addr] + 256 * mem[dis.addr + 1];
		len = pad_argument("%s", get_sym(dis.ind16, SYM_NONE));
		break;
	case _dIMPL:
		dis.flag.dst_addr = 1;
		dis.addr  = (dis.cmd0 >> 3) & 7;
		len = pad_argument("%s", hexb(dis.addr));
		break;
	}
	if (comma) {
		sprintf(dis.dst, ",");
		dis.dst++;
		len++;
	}
	return len;
}

/***************************************************************************** */
/* disassemble a source operand from the given table entry */
/*******************************************************************---------- */
int     dis_src(TABLE * ptab)
{
	int	len = 0;

	switch (ptab->src) {
	case _s0:
		return 0;

	case _sB:
		len = pad_argument("%s","B");
		dis.val8 = i8085.bc.r8[1];
		break;

	case _sC:
		len = pad_argument("%s","C");
		dis.val8 = i8085.bc.r8[0];
		break;

	case _sD:
		len = pad_argument("%s","D");
		dis.val8 = i8085.de.r8[1];
		break;

	case _sE:
		len = pad_argument("%s","E");
		dis.val8 = i8085.de.r8[0];
		break;

	case _sH:
		len = pad_argument("%s","H");
		dis.val8 = i8085.hl.r8[1];
		break;

	case _sL:
		len = pad_argument("%s","L");
		dis.val8 = i8085.hl.r8[0];
		break;

	case _sM:
		dis.flag.src_mem8 = 1;
		dis.flag.src_addr = 1;
		dis.addr  = i8085.hl.r16;
		dis.mem8  = dis.val8 = mem[dis.addr];
		len = pad_argument("%s","M");
		break;

	case _sA:
		len = pad_argument("%s","A");
		dis.val8 = i8085.af.r8[1];
		break;

	case _sIM:
		len = pad_argument("%s","IM");
		dis.val8 = i8085.im;
		break;

	case _sMBC:
		dis.flag.src_mem8 = 1;
		dis.flag.src_addr = 1;
		dis.addr  = i8085.bc.r16;
		dis.mem8  = dis.val8 = mem[dis.addr];
		len = pad_argument("%s","B");
		break;

	case _sMDE:
		dis.flag.src_mem8 = 1;
		dis.flag.src_addr = 1;
		dis.addr  = i8085.de.r16;
		dis.mem8  = dis.val8 = mem[dis.addr];
		len = pad_argument("%s","D");
		break;

	case _sMSP:
		dis.flag.src_mem16 = 1;
		dis.flag.src_addr  = 1;
		dis.addr  = i8085.sp.r16;
		dis.mem16 = mem[dis.addr] + 256 * mem[dis.addr + 1];
		len = pad_argument("%s","(SP)");
		break;

	case _sIND8:
		if (!dis.flag.has_ind) {
			dis.flag.has_ind = 1;
			dis.ind16 = mem[dis.pc++];
			dis.ind16 += 256 * mem[dis.pc++];
		}
		dis.flag.src_mem8 = 1;
		dis.addr  = dis.ind16;
		dis.mem8  = dis.val8 = mem[dis.addr];
		len = pad_argument("%s", get_sym(dis.ind16, SYM_NONE));
		break;

	case _sIMM8:
		if (!dis.flag.has_imm8) {
			dis.flag.has_imm8 = 1;
			dis.imm8  = mem[dis.pc++];
		}
		dis.val8 = dis.imm8;
		len = pad_argument("%s", hexb(dis.imm8));
		break;

	case _sIMM8A:
		if (!dis.flag.has_imm8) {
			dis.flag.has_imm8 = 1;
			dis.imm8  = mem[dis.pc++];
		}
		dis.val8 = dis.imm8;
		switch (dis.imm8) {
			case 0 ... 31:
				len = pad_argument("%d", dis.imm8);
				break;
			case 32 ... 127:
				len = pad_argument("\"%c\"", dis.imm8);
				break;
			case 255:
				len = pad_argument("-1");
				break;
			default:
				len = pad_argument("%s", hexb(dis.imm8));
				break;
		}
		break;

	case _sPORT8:
		if (!dis.flag.has_port) {
			dis.flag.has_port = 1;
			dis.port  = mem[dis.pc++];
		}
		dis.flag.src_port = 1;
		dis.val8 = dis.port;
		len = pad_argument("%s", hexb(dis.port));
		break;

	case _sBC:
		len = pad_argument("%s","B");
		dis.val16 = i8085.bc.r16;
		break;

	case _sDE:
		len = pad_argument("%s","D");
		dis.val16 = i8085.de.r16;
		break;

	case _sHL:
		len = pad_argument("%s","H");
		dis.val16 = i8085.hl.r16;
		break;

	case _sSP:
		len = pad_argument("%s","SP");
		dis.val16 = i8085.sp.r16;
		break;

	case _sAF:
		len = pad_argument("%s","PSW");
		dis.val16 = i8085.af.r16;
		break;

	case _sPC:
		len = pad_argument("%s","$");
		break;

	case _sPCHL:
		dis.flag.src_addr = 1;
		dis.addr = dis.val16 = i8085.hl.r16;
		len = pad_argument("%s","HL");
		break;

	case _sIMM16:
		if (!dis.flag.has_imm16) {
			dis.flag.has_imm16 = 1;
			dis.imm16 = mem[dis.pc++];
			dis.imm16 += 256 * mem[dis.pc++];
		}
		dis.val16 = dis.imm16;
		len = pad_argument("%s", get_sym(dis.imm16, SYM_NONE));
		break;

	case _sADR16:
		if (!dis.flag.has_adr16) {
			dis.flag.has_adr16 = 1;
			dis.adr16 = mem[dis.pc++];
			dis.adr16 += 256 *mem[dis.pc++];
		}
		dis.flag.jmp_addr = 1;
		dis.addr = dis.val16 = dis.adr16;
		len = pad_argument("%s", get_sym(dis.adr16, SYM_NONE));
		break;

	case _sIND16:
		if (!dis.flag.has_ind) {
			dis.flag.has_ind = 1;
			dis.ind16 = mem[dis.pc++];
			dis.ind16 += 256 * mem[dis.pc++];
		}
		dis.flag.src_mem16 = 1;
		dis.addr  = dis.ind16;
		dis.mem16 = dis.val16 = mem[dis.addr] + 256 * mem[dis.addr + 1];
		len = pad_argument("%s", get_sym(dis.ind16, SYM_NONE));
		break;

	case _sIMPL:
		dis.flag.jmp_addr = 1;
		dis.addr  = (dis.cmd0 >> 3) & 7;
		len = pad_argument("%s", hexb(dis.addr));
		break;
	}
	return len;
}

/***************************************************************************** */
/* disassemble one instruction from the given table entry */
/*******************************************************************---------- */
void    dis_instr(TABLE * ptab)
{
	switch (ptab->oper) {
	case _t0:
		pad_opcode("???");
		break;
		
	case _tADD:
		pad_opcode("ADD");
		dis_src(ptab);
		dis.val8  += i8085.af.r8[1];
		break;
		
	case _tADI:
		pad_opcode("ADI");
		dis_src(ptab);
		dis.val8  += i8085.af.r8[1];
		break;
		
	case _tADC:
		pad_opcode("ADC");
		dis_src(ptab);
		dis.val8  += i8085.af.r8[1] + i8085.af.f.c;
		break;
		
	case _tACI:
		pad_opcode("ACI");
		dis_src(ptab);
		dis.val8  += i8085.af.r8[1] + i8085.af.f.c;
		break;
		
	case _tANA:
		pad_opcode("ANA");
		dis_src(ptab);
		dis.val8 = (i8085.af.r8[1] & dis.val8);
		break;
		
	case _tANI:
		pad_opcode("ANI");
		dis_src(ptab);
		dis.val8 = (i8085.af.r8[1] & dis.val8);
		break;

	case _tCNZ:
		pad_opcode("CNZ");
		dis_src(ptab);
		break;

	case _tCZ:
		pad_opcode("CZ");
		dis_src(ptab);
		break;

	case _tCNC:
		pad_opcode("CNC");
		dis_src(ptab);
		break;

	case _tCC:
		pad_opcode("CC");
		dis_src(ptab);
		break;

	case _tCPO:
		pad_opcode("CPO");
		dis_src(ptab);
		break;

	case _tCPE:
		pad_opcode("CPE");
		dis_src(ptab);
		break;

	case _tCP:
		pad_opcode("CP");
		dis_src(ptab);
		break;

	case _tCM:
		pad_opcode("CM");
		dis_src(ptab);
		break;
		
	case _tCALL:
		pad_opcode("CALL");
		dis_src(ptab);
		break;

	case _tCCF:
		pad_opcode("CCF");
		i8085.af.f.c = ~i8085.af.f.c;
		break;

	case _tCMA:
		pad_opcode("CMA");
		dis.val8 = ~i8085.af.r8[1];
		break;

	case _tCMP:
		pad_opcode("CMP");
		dis_src(ptab);
		break;

	case _tCPI:
		pad_opcode("CPI");
		dis_src(ptab);
		break;

	case _tDAA:
		pad_opcode("DAA");
		break;

	case _tDAD:
		pad_opcode("DAD");
		dis_src(ptab);
		dis.val16 += i8085.hl.r16;
		break;

	case _tDCX:
		pad_opcode("DCX");
		dis_src(ptab);
		dis.val16 -= 1;
		break;

	case _tDCR:
		pad_opcode("DCR");
		dis_src(ptab);
		dis.val8  -= 1;
		break;

	case _tDI:
		pad_opcode("DI");
		break;

	case _tEI:
		pad_opcode("EI");
		break;

	case _tHALT:
		pad_opcode("HALT");
		break;

	case _tIN:
		pad_opcode("IN");
		dis_src(ptab);
		break;

	case _tINX:
		pad_opcode("INX");
		dis_src(ptab);
		dis.val16 += 1;
		break;

	case _tINR:
		pad_opcode("INR");
		dis_src(ptab);
		dis.val8  += 1;
		break;

	case _tJNZ:
		pad_opcode("JNZ");
		dis_src(ptab);
		break;

	case _tJZ:
		pad_opcode("JZ");
		dis_src(ptab);
		break;

	case _tJNC:
		pad_opcode("JNC");
		dis_src(ptab);
		break;

	case _tJC:
		pad_opcode("JC");
		dis_src(ptab);
		break;

	case _tJPO:
		pad_opcode("JPO");
		dis_src(ptab);
		break;

	case _tJPE:
		pad_opcode("JPE");
		dis_src(ptab);
		break;

	case _tJP:
		pad_opcode("JP");
		dis_src(ptab);
		break;

	case _tJM:
		pad_opcode("JM");
		dis_src(ptab);
		break;

	case _tJMP:
		pad_opcode("JMP");
		dis_src(ptab);
		dis.flag.end_mode = 1;
		break;

	case _tMOV:
		pad_opcode("MOV");
		dis_dst(ptab, 1);
		dis_src(ptab);
		break;

	case _tMVI:
		pad_opcode("MVI");
		dis_dst(ptab, 1);
		dis_src(ptab);
		break;

	case _tLXI:
		pad_opcode("LXI");
		dis_dst(ptab, 1);
		dis_src(ptab);
		break;

	case _tLDA:
		pad_opcode("LDA");
		dis_src(ptab);
		break;

	case _tSTA:
		pad_opcode("STA");
		dis_dst(ptab, 0);
		break;

	case _tLDAX:
		pad_opcode("LDAX");
		dis_src(ptab);
		break;

	case _tSTAX:
		pad_opcode("STAX");
		dis_dst(ptab, 0);
		break;

	case _tLHLD:
		pad_opcode("LHLD");
		dis_src(ptab);
		break;

	case _tSHLD:
		pad_opcode("SHLD");
		dis_dst(ptab, 0);
		break;

	case _tNOP:
		pad_opcode("NOP");
		break;

	case _tORA:
		pad_opcode("ORA");
		dis_src(ptab);
		dis.val8 |= i8085.af.r8[1];
		break;

	case _tORI:
		pad_opcode("ORI");
		dis_src(ptab);
		dis.val8 |= i8085.af.r8[1];
		break;

	case _tOUT:
		pad_opcode("OUT");
		dis_dst(ptab, 0);
		break;

	case _tPOP:
		pad_opcode("POP");
		dis_dst(ptab, 0);
		break;

	case _tPUSH:
		pad_opcode("PUSH");
		dis_src(ptab);
		break;

	case _tPCHL:
		pad_opcode("PCHL");
		dis.pc = i8085.hl.r16;
		dis.flag.end_mode = 1;
		break;

	case _tSPHL:
		pad_opcode("SPHL");
		break;

	case _tRNZ:
		pad_opcode("RNZ");
		break;

	case _tRZ:
		pad_opcode("RZ");
		break;

	case _tRNC:
		pad_opcode("RNC");
		break;

	case _tRC:
		pad_opcode("RC");
		break;

	case _tRPO:
		pad_opcode("RPO");
		break;

	case _tRPE:
		pad_opcode("RPE");
		break;

	case _tRP:
		pad_opcode("RP");
		break;

	case _tRM:
		pad_opcode("RM");
		break;

	case _tRET:
		pad_opcode("RET");
		dis.flag.end_mode = 1;
		break;

	case _tRAL:
		pad_opcode("RAL");
		dis.val8 = (i8085.af.r8[1] << 1) | (i8085.af.f.c);
		break;

	case _tRAR:
		pad_opcode("RAR");
		dis.val8 = (i8085.af.r8[1] >> 1) | (i8085.af.f.c << 7);
		break;

	case _tRLC:
		pad_opcode("RLC");
		dis.val8 = (i8085.af.r8[1] << 1) | (i8085.af.r8[1] >> 7);
		break;

	case _tRRC:
		pad_opcode("RRC");
		dis.val8 = (i8085.af.r8[1] >> 1) | (i8085.af.r8[1] << 7);
		break;

	case _tRST:
		pad_opcode("RST");
		dis_src(ptab);
		break;

	case _tSBB:
		pad_opcode("SBB");
		dis_src(ptab);
		dis.val8  = (i8085.af.r8[1] - dis.val8 - i8085.af.f.c);
		break;

	case _tSBI:
		pad_opcode("SBI");
		dis_src(ptab);
		dis.val8  = (i8085.af.r8[1] - dis.val8 - i8085.af.f.c);
		break;

	case _tSTC:
		pad_opcode("STC");
		i8085.af.f.c = 1;
		break;

	case _tSUB:
		pad_opcode("SUB");
		dis_src(ptab);
		dis.val8 = (i8085.af.r8[1] - dis.val8);
		break;

	case _tSUI:
		pad_opcode("SUI");
		dis_src(ptab);
		dis.val8 = (i8085.af.r8[1] - dis.val8);
		break;

	case _tXRA:
		pad_opcode("XRA");
		dis_src(ptab);
		dis.val8 = (i8085.af.r8[1] ^ dis.val8);
		break;

	case _tXRI:
		pad_opcode("XRI");
		dis_src(ptab);
		dis.val8 = (i8085.af.r8[1] ^ dis.val8);
		break;

	case _tXCHG:
		pad_opcode("XCHG");
		dis.val16 = i8085.de.r16;
		i8085.de.r16 = i8085.hl.r16;
		i8085.hl.r16 = dis.val16;
		break;

	case _tXTHL:
		pad_opcode("XTHL");
		break;

	case _tRIM:
		pad_opcode("RIM");
		break;

	case _tSIM:
		pad_opcode("SIM");
		break;

	case _tDSUB:
		pad_opcode("DSUB");
		dis_src(ptab);
		break;

	case _tARHL:
		pad_opcode("ARHL");
		break;

	case _tRLDE:
		pad_opcode("RLDE");
		break;

	case _tLDHI:
		pad_opcode("LDHI");
		dis_src(ptab);
		break;

	case _tLDSI:
		pad_opcode("LDSI");
		dis_src(ptab);
		break;

	case _tLHLX:
		pad_opcode("LHLX");
		dis_src(ptab);
		break;

	case _tSHLX:
		pad_opcode("SHLX");
		dis_dst(ptab, 0);
		break;

	case _tRSTV:
		pad_opcode("RSTV");
		break;

	case _tJNX:
		pad_opcode("JNX");
		dis_src(ptab);
		break;

	case _tJX:
		pad_opcode("JX");
		dis_src(ptab);
		break;

	case _tEXT:
		break;
	}
}

char *dis_code(u16 * addr)
{
	TABLE   * ptab;

	dis.pc   = *addr;
	dis.cmd0  = mem[dis.pc++];
	dis.dst   = dis.buff + COLUMN_OPCODE;

	ptab = dis_xx + dis.cmd0;
	dis_instr(ptab);

	if (!ptab->oper)
		dis.dst += pad_opcode("???  %02X", dis.cmd0);

	*dis.dst   = '\0';

	if (dis.prtcom) {
		dis.prtcom = 0;
		pad_comment(" %s", com_get(*addr, COM_APPEND));
	} else {
		if (dis.flag.src_mem8) {
			pad_comment(" [%04X] {%02X}",
				dis.addr, dis.mem8);
			dis.flag.src_addr = 0;
		} else if (dis.flag.dst_mem8) {
			pad_comment(" [%04X] {%02X}",
				dis.addr, dis.mem8);
			dis.flag.dst_addr = 0;
		}

		if (dis.flag.src_mem16)
		{
			pad_comment(" [%04X] {%s}",
				dis.addr, get_sym(dis.mem16, SYM_NONE));
			dis.flag.src_addr = 0;
		} else if (dis.flag.dst_mem16) {
			pad_comment(" [%04X] {%s}",
				dis.addr, get_sym(dis.mem16, SYM_NONE));
			dis.flag.dst_addr = 0;
		}

		if (dis.flag.src_port) {
			pad_comment(" (%04X)", dis.port);
		}

		if (dis.flag.dst_port) {
			pad_comment(" (%04X)", dis.port);
		}

		if (dis.flag.src_addr) {
			pad_comment(" %04X", dis.addr);
		} else if (dis.flag.dst_addr) {
			pad_comment(" %04X", dis.addr);
		} else if (dis.flag.jmp_addr) {
			pad_comment(" %04X", dis.addr);
		}
	}

	sprintf(dis.buff, "%04X:%02X ", *addr, dis.cmd0);

	dis.dst = dis.buff + 8;
	*addr  += 1;

	if (dis.flag.has_cmd1) {
		dis.dst += sprintf(dis.dst, "%02X ", dis.cmd1);
		*addr	+= 1;
	}

	if (dis.flag.has_port) {
		dis.dst += sprintf(dis.dst, "%02X ", dis.port);
		*addr	+= 1;
	}

	if (dis.flag.has_imm8) {
		dis.dst += sprintf(dis.dst, "%02X ", dis.imm8);
		*addr	+= 1;
	}

	if (dis.flag.has_ofs8) {
		dis.dst += sprintf(dis.dst, "%02X ", (u8)dis.ofs8);
		*addr	+= 1;
	}

	if (dis.flag.has_imm16) {
		dis.dst += sprintf(dis.dst, "%04X ", dis.imm16);
		*addr	+= 2;
	}

	if (dis.flag.has_adr16) {
		dis.dst += sprintf(dis.dst, "%04X ", dis.adr16);
		*addr	+= 2;
	}

	if (dis.flag.has_ind) {
		dis.dst += sprintf(dis.dst, "%04X ", dis.ind16);
		*addr	+= 2;
	}

	while (dis.dst < dis.buff + COLUMN_OPCODE)
		*dis.dst++ = ' ';

	switch (ptab->dst) {
	case _d0:
	case _dM:
	case _dMBC:
	case _dMDE:
	case _dMSP:
	case _dIMM8:
	case _dIND8:
	case _dPC:
	case _dIND16:
	case _dIMPL:
		break;
	case _dB:
		i8085.bc.r8[1] = dis.val8;
		break;
	case _dC:
		i8085.bc.r8[0] = dis.val8;
		break;
	case _dD:
		i8085.de.r8[1] = dis.val8;
		break;
	case _dE:
		i8085.de.r8[0] = dis.val8;
		break;
	case _dH:
		i8085.hl.r8[1] = dis.val8;
		break;
	case _dL:
		i8085.hl.r8[0] = dis.val8;
		break;
	case _dA:
		i8085.af.r8[1] = dis.val8;
		break;
	case _dIM:
		i8085.im = dis.val8;
		break;
	case _dPORT8:
		break;
	case _dBC:
		i8085.bc.r16 = dis.val16;
		break;
	case _dDE:
		i8085.de.r16 = dis.val16;
		break;
	case _dHL:
		i8085.hl.r16 = dis.val16;
		break;
	case _dSP:
		i8085.sp.r16 = dis.val16;
		break;
	case _dAF:
		i8085.af.r16 = dis.val16;
		break;
	case _dPCHL:
		i8085.pc.r16 = dis.val16;
		break;
	}

	if (dis.flag.end_mode)
		memset(&i8085, 0, sizeof(i8085));

	return dis.buff;
}


int dis_bytes(void)
{
	int     i;
	dis.dst += pad_opcode("DEFB");
	for (i = 0; i < 8; i++) {
		if (i > 0) {
			if (memopt[dis.pc].m.symbol)
				break;

			if (memopt[dis.pc].m.comment)
				break;

			if (NEWOPT(dis.pc))
				break;

			dis.dst += pad_argument(",");
		}
		dis.val8 = mem[dis.pc++];
		dis.dst += pad_argument("%s", hexb(dis.val8));
	}
	return i;
}

int dis_ascii(void)
{
	u8	str = 0;
	u8    old = 0;
	int     i;

	dis.dst += pad_opcode("DEFM");
	for (i = 0; i < 8; i++) {
		if (i > 0)
		{
			if ((dis.val8 == 0x00) && (old != 0x00))
				break;

			if (memopt[dis.pc].m.symbol)
				break;

			if (memopt[dis.pc].m.comment)
				break;

			if (NEWOPT(dis.pc))
				break;

			if (!str)
				dis.dst += pad_argument(",");
		}
		old = dis.val8;
		dis.val8 = mem[dis.pc++];
		switch (dis.val8) {
		case 0x00 ... 0x1F:
		case 0x22:
			if (str)
				dis.dst += pad_argument("\",");
			dis.dst += pad_argument("%s", hexb(dis.val8));
			str = 0;
			break;
		default:
			if (!str)
				dis.dst += pad_argument("\"");
			dis.dst += pad_argument("%c", dis.val8);
			str = 1;
			break;
		}
	}
	if (str)
		dis.dst += pad_argument("\"");
	return i;
}

int dis_7bit(void)
{
	int     first_bit8;
	u8    str = 0;
	int     i;

	dis.dst += pad_opcode("DEFM");
	first_bit8 = (mem[dis.pc] & 0x80) ? 1 : 0;
	for (i = 0; i < 8; i++) {
		if (i > 0) {
			if (mem[dis.pc] == 0x80) {
				dis.flag.end_mode = 1;
				break;
			}

			if (first_bit8 && (mem[dis.pc] & 0x80))
				break;

			if (!first_bit8 && (dis.val8 & 0x80))
				break;

			if (memopt[dis.pc].m.symbol)
				break;

			if (memopt[dis.pc].m.comment)
				break;

			if (NEWOPT(dis.pc))
				break;

			if (!str)
				dis.dst += pad_argument(",");
		}
		dis.val8 = mem[dis.pc++];
		if (dis.val8 < 0x20 ||
			dis.val8 == 0x22 ||
			(dis.val8 >= 0x80 && dis.val8 < 0xa0) ||
			dis.val8 == 0xA7) {
			if (str)
				dis.dst += pad_argument("\",");
			dis.dst += pad_argument("%s", hexb(dis.val8));
			str = 0;
		} else if (dis.val8 >= 0xa0) {
			if (str)
				dis.dst += pad_argument("\",");
			dis.dst += pad_argument("'%c'+128",
				dis.val8 & 0x7F);
			str = 0;
		} else {
			if (!str)
				dis.dst += pad_argument("\"");
			dis.dst += pad_argument("%c", dis.val8);
			str = 1;
		}
	}
	if (str)
		dis.dst += pad_argument("\"");
	return i;
}

int dis_words(void)
{
	int	i;
	dis.dst += pad_opcode("DEFW");
	for (i = 0; i < 4; i++) {
		if (i > 0) {
			if (memopt[dis.pc].m.symbol)
				break;

			if (memopt[dis.pc].m.comment)
				break;

			if (NEWOPT(dis.pc))
				break;

			dis.dst += pad_argument(",");
		}
		dis.val16 = mem[dis.pc++];
		dis.val16 += 256 * mem[dis.pc++];
		dis.dst += pad_argument("%s", hexw(dis.val16));
	}
	return i;
}

int	dis_entry(void)
{
	dis.dst += pad_opcode("DEFW");
	dis.val16 = mem[dis.pc++];
	dis.val16 += 256 * mem[dis.pc++];
	dis.dst += pad_argument("%s", get_sym(dis.val16, SYM_NONE));
	return 1;
}

double	frac2[56] = {
	0.5,
	0.25,
	0.125,
	0.0625,
	0.03125,
	0.015625,
	0.0078125,
	0.00390625,
	0.001953125,
	0.0009765625,
	0.00048828125,
	0.000244140625,
	0.0001220703125,
	0.00006103515625,
	0.000030517578125,
	0.0000152587890625,
	0.00000762939453125,
	0.000003814697265625,
	0.0000019073486328125,
	0.00000095367431640625,
	0.000000476837158203125,
	0.0000002384185791015625,
	0.00000011920928955078125,
	0.000000059604644775390625,
	0.0000000298023223876953125,
	0.00000001490116119384765625,
	0.000000007450580596923828125,
	0.0000000037252902984619140625,
	0.00000000186264514923095703125,
	0.000000000931322574615478515625,
	0.0000000004656612873077392578125,
	0.00000000023283064365386962890625,
	0.000000000116415321826934814453125,
	0.0000000000582076609134674072265625,
	0.00000000002910383045673370361328125,
	0.000000000014551915228366851806640625,
	0.0000000000072759576141834259033203125,
	0.00000000000363797880709171295166015625,
	0.000000000001818989403545856475830078125,
	0.0000000000009094947017729282379150390625,
	0.00000000000045474735088646411895751953125,
	0.000000000000227373675443232059478759765625,
	0.0000000000001136868377216160297393798828125,
	0.00000000000005684341886080801486968994140625,
	0.000000000000028421709430404007434844970703125,
	0.0000000000000142108547152020037174224853515625,
	0.00000000000000710542735760100185871124267578125,
	0.000000000000003552713678800500929355621337890625,
	0.0000000000000017763568394002504646778106689453125,
	0.00000000000000088817841970012523233890533447265625,
	0.000000000000000444089209850062616169452667236328125,
	0.0000000000000002220446049250313080847263336181640625,
	0.00000000000000011102230246251565404236316680908203125,
	0.000000000000000055511151231257827021181583404541015625,
	0.0000000000000000277555756156289135105907917022705078125,
	0.00000000000000001387778780781445675529539585113525390625,
};

int dis_float_single(void)
{
	u8 b[4];
	int i;
	double df;

	b[0] = mem[dis.pc++];
	b[1] = mem[dis.pc++];
	b[2] = mem[dis.pc++];
	b[3] = mem[dis.pc++];
	dis.fsng.man  = b[0];
	dis.fsng.man += b[1] << 8;
	dis.fsng.man += b[2] << 16;
	dis.fsng.exp  = b[3] - 128;
	dis.dst += pad_opcode("DEFB");
	dis.dst += pad_argument("%s,%s,%s,%s",
		hexb(b[0]), hexb(b[1]), hexb(b[2]), hexb(b[3]));

	if (dis.prtcom)
		return 1;

	df = frac2[0];
	for (i = 1; i < 24; i++)
		if (dis.fsng.man & (0x800000 >> i))
			df += frac2[i];

	if (dis.fsng.exp > 0)
		df *= 1 << dis.fsng.exp;
	else
		df /= 1 << -dis.fsng.exp;

	if (dis.fsng.man & 0x800000)
		df = -df;

	dis.dst += pad_comment(" %g", df);

	return 1;
}

int dis_float_double(void)
{
	u8 b[8];
	int i;
	double df;

	b[0] = mem[dis.pc++];
	b[1] = mem[dis.pc++];
	b[2] = mem[dis.pc++];
	b[3] = mem[dis.pc++];
	b[4] = mem[dis.pc++];
	b[5] = mem[dis.pc++];
	b[6] = mem[dis.pc++];
	b[7] = mem[dis.pc++];
	dis.fdbl.manl  = b[0];
	dis.fdbl.manl += b[1] << 8;
	dis.fdbl.manl += b[2] << 16;
	dis.fdbl.manl += b[3] << 24;
	dis.fdbl.manh  = b[4];
	dis.fdbl.manh += b[5] << 8;
	dis.fdbl.manh += b[6] << 16;
	dis.fdbl.exp   = b[7] - 128;
	dis.dst += pad_opcode("DEFB");
	dis.dst += pad_argument("%s,%s,%s,%s,%s,%s,%s,%s",
		hexb(b[0]), hexb(b[1]), hexb(b[2]), hexb(b[3]),
		hexb(b[4]), hexb(b[5]), hexb(b[6]), hexb(b[7]));

	if (dis.prtcom)
		return 1;

	df = frac2[0];
	for (i = 1; i < 32; i++)
		if (dis.fdbl.manl & (0x80000000 >> i))
			df += frac2[i];
	for (i = 0; i < 24; i++)
		if (dis.fdbl.manh & (0x800000 >> i))
			df += frac2[32+i];
	if (dis.fdbl.exp > 0)
		df *= 1 << dis.fdbl.exp;
	else
		df /= 1 << -dis.fdbl.exp;

	if (dis.fdbl.manh & 0x800000)
		df = -df;

	dis.dst += pad_comment(" %g", df);
	return 1;
}

char *dis_data(u16 * addr)
{
	char *phex;
	int cnt;

	dis.pc  = *addr;

	dis.dst = dis.buff + sprintf(dis.buff, "%04X:", dis.pc);

	phex	= dis.dst;

	switch (dis.curopt.m.opt) {
	case OPT_BYTES:
		cnt = dis_bytes();
		break;

	case OPT_ASCII:
		cnt = dis_ascii();
		break;

	case OPT_7BIT:
		cnt = dis_7bit();
		break;

	case OPT_WORDS:
		cnt = dis_words();
		break;

	case OPT_ENTRY:
		cnt = dis_entry();
		break;

	case OPT_FSNG:
		cnt = dis_float_single();
		break;

	case OPT_FDBL:
		cnt = dis_float_double();
		break;

	default:
		cnt = dis_bytes();
		break;
	}

	dis.pc  = *addr;

	while (cnt--) {
		switch (dis.curopt.m.size) {
		case 0:
			dis.val8 = mem[dis.pc++];
			phex += sprintf(phex, "%02X ", dis.val8);
			break;
		case 1:
			dis.val16  = mem[dis.pc++];
			dis.val16 += mem[dis.pc++] << 8;
			phex += sprintf(phex, "%04X ", dis.val16);
			break;
		case 2:
			dis.fsng.man  = mem[dis.pc++];
			dis.fsng.man += mem[dis.pc++] << 8;
			dis.fsng.man += mem[dis.pc++] << 16;
			dis.fsng.exp  = mem[dis.pc++];
			phex += sprintf(phex, "%06X %02X",
				dis.fsng.man, (unsigned) dis.fsng.exp);
			break;
		case 3:
			dis.fdbl.manl = mem[dis.pc++];
			dis.fdbl.manl+= mem[dis.pc++] << 8;
			dis.fdbl.manl+= mem[dis.pc++] << 16;
			dis.fdbl.manl+= mem[dis.pc++] << 24;
			dis.fdbl.manh = mem[dis.pc++];
			dis.fdbl.manh+= mem[dis.pc++] << 8;
			dis.fdbl.manh+= mem[dis.pc++] << 16;
			dis.fdbl.exp  = mem[dis.pc++] - 128;
			phex += sprintf(phex, "%06X%08X %02X ",
				dis.fdbl.manh, dis.fdbl.manl,
				(unsigned) dis.fdbl.exp);
			break;
		}
	}

	*phex = ' ';

	if (dis.prtcom) {
		dis.prtcom = 0;
		dis.dst += pad_comment(" %s", com_get(*addr, COM_APPEND));
	}

	*addr   = dis.pc;

	return dis.buff;
}

void    disasm_init(u16 addr)
{
	memset(&i8085, 0, sizeof(i8085));

	dis.curopt = memopt[addr];

	if (dis.curopt.m.comment) {
		dis.curopt.m.comment = 0;
		dis.prtcom = 1;
	}

	if (dis.curopt.m.symbol) {
		dis.curopt.m.symbol  = 0;
		dis.prtsym = 1;
	}
}

char *disasm(u16 * addr, int flg_dump)
{
	char *src;
	char *dst = "";
	u16 addr_cmd;

	addr_cmd = *addr;

	memset(&dis.flag, 0, sizeof(dis.flag));
	dis.padcom = 0;

	if (dis.prtcom) {
		dis.prtcom = 0;
		src = com_get(*addr, COM_BLOCK);
		if (src && strlen(src))
		{
			dst = dis.buff;
			sprintf(dst, ";****** %s", com_get(addr_cmd, COM_BLOCK));
			return dst;
		}
	}

	if (flg_dump) {
		if (dis.prtsym) {
			dis.prtsym = 0;
			dst = dis.buff;
			if (dis.curopt.m.mode == MEM_CODE)
				sprintf(dst, "%s:",
					get_sym(addr_cmd, SYM_NONE));
			else
				sprintf(dst, "%-*sEQU  $", COLUMN_OPCODE,
					get_sym(addr_cmd, SYM_NONE));
			return dst;
		}
	}

	switch (dis.curopt.m.mode) {
	case MEM_UNKNOWN:
		dst = dis_data(addr);
		break;

	case MEM_CODE:
		dst = dis_code(addr);
		break;

	case MEM_DATA:
		dst = dis_data(addr);
		break;

	case MEM_IO:
		dst = dis_data(addr);
		break;
	}

	if (!flg_dump) {
		if (dis.prtsym) {
			dis.prtsym = 0;
			dst[sprintf(dst, "%-*s ", COLUMN_OPCODE-2, get_sym(addr_cmd, SYM_NONE))] = ' ';
		} else {
			dst[sprintf(dst, "%-*s", COLUMN_OPCODE-1, "")] = ' ';
		}
	}

	if (NEWOPT(*addr)) {
		u8 old_mode = dis.curopt.m.mode;
		u8 old_opt  = dis.curopt.m.opt;

		dis.curopt = memopt[*addr];

		if (dis.curopt.m.mode == MEM_UNKNOWN) {
			dis.curopt.m.mode = old_mode;
			if (dis.curopt.m.opt  == OPT_NONE)
				dis.curopt.m.opt = old_opt;
		}

		if (dis.curopt.m.comment) {
			dis.curopt.m.comment = 0;
			dis.prtcom = 1;
		}

		if (dis.curopt.m.symbol) {
			dis.curopt.m.symbol  = 0;
			dis.prtsym = 1;
		}
	}

	if (dis.prtsym || dis.flag.end_mode)
		strcat(dst, "\n");

	return dst;
}
