typedef	struct {
	int has_port:1;
	int has_imm16:1;
	int has_adr16:1;
	int has_ind:1;
	int has_cmd1:1;
	int has_cmd3:1;
	int has_imm8:1;
	int has_ofs8:1;
	int has_rel8:1;
	int end_mode:1;
	int src_mem8:1;
	int dst_mem8:1;
	int src_mem16:1;
	int dst_mem16:1;
	int src_port:1;
	int dst_port:1;
	int dst_addr:1;
	int src_addr:1;
	int jmp_addr:1;
} disflag_t;

#define MEM_UNKNOWN  0
#define MEM_CODE     1
#define MEM_DATA     2
#define MEM_IO	     3

typedef union {
	u32 w;
	struct {
		/* memory contains 0:unknown, 1:code, 2:data, 3:io */
		unsigned mode:2;
		/* memory location has a symbol */
	        unsigned symbol:1;
		/* memory location has a comment */
	        unsigned comment:1;
		/* size of data elements (1,2,4,8) */
	        unsigned size:2;
		/* data options */
		unsigned opt:10;
	} m;
} memopt_t;

#define OPT_NONE	0x000	/* no special memory options */
#define OPT_BYTES	0x001	/* memory contains byte data */
#define OPT_ASCII	0x002	/* memory contains ascii data */
#define OPT_7BIT	0x003	/* memory contains ascii data with opt. bit 7 */
#define OPT_WORDS	0x008	/* memory contains (16 bit) word data */
#define OPT_ENTRY	0x009	/* memory contains (16 bit) entry points */
#define OPT_FSNG	0x00A	/* memory contains single float (4 bytes) */
#define OPT_FDBL	0x00B	/* memory contains double float (8 bytes) */

#define OPT_OC		0x010	/* memory contains offset to code */
#define OPT_OD7 	0x011	/* memory contains offset to data 7bit */
#define OPT_ODA 	0x012	/* memory contains offset to data ascii */
#define OPT_ODB 	0x013	/* memory contains offset to data bytes */
#define OPT_ODW 	0x014	/* memory contains offset to data words */
#define OPT_ODE 	0x015	/* memory contains offset to data entry */

#define OPT_OFFS_CODE	0x100	/* immediate is offset to code */
#define OPT_OFFS_DATA	0x101	/* immediate is offset to data */

#define OPT_USER	0x200	/* user options 0x200-0x3FF */

typedef struct {
	unsigned man:24;
	signed exp:8;
}	float_sng_t;

typedef struct	tagFDBL {
	unsigned manl:32;
	unsigned manh:24;
	signed exp:8;
}	float_dbl_t;

typedef struct {
        char buff[8192];
	char name[32];
	char *dst;
	disflag_t flag;		/* flags after disassembly */
	u16 pc;			/* offset into mem */
	u16 port;		/* port address */
	u16 imm16;		/* immediate 16bit argument */
	u16 adr16;		/* jump address 16bit */
	u16 ind16;		/* indirect 16bit memory address */
	u8 cmd0;		/* command byte 0 */
	u8 cmd1;		/* command byte 1 (at CB, DD, ED and FD) */
	u8 cmd3;		/* command byte 3 (only at DD CB oo XX and FD CB oo XX) */
	u8 imm8;		/* imediate 8bit argument */
	i8 ofs8;		/* IX or IY relative offset */
	i8 rel8;		/* PC relative offset */
	u8 mem8;		/* 8 bit mem contents */
	u8 val8;		/* 8 bit transfer value */
	u16 mem16;		/* 16 bit mem contents */
	u16 val16;		/* 16 bit transfer value */
        u16 addr;		/* jump/call address */
	float_sng_t fsng;	/* single precision float (32 bits) */
	float_dbl_t fdbl;	/* double precision float (64 bits) */
	memopt_t curopt;	/* current options */
        int prtcom;		/* true if comment is to be printed out */
        int prtsym;		/* true if symbol is to be printed out */
	int padcom;		/* true if comment semicolon (;) already printed */
}	dis_t;

typedef	enum {
	SYM_NONE=0,
        SYM_CODE,
	SYM_DATA,
	SYM_PORT
} symtype_t;

typedef struct sym_s {
	struct sym_s * next;
	symtype_t type;
	u16 addr;
        char name[1];
} sym_t;

typedef enum {
	COM_NONE=0,
	COM_BLOCK,
	COM_APPEND,
} comtype_t;

typedef struct com_s {
	struct com_s *next;
	comtype_t type;
	u16 addr;
	char text[1];
} com_t;

void def_opt(u16 addr, unsigned mode, unsigned size, unsigned opt);
void add_sym(u16 addr, symtype_t type, char * name);
void com_t_add(u16 addr, comtype_t type, char * name);
char *get_sym(u16 addr, symtype_t type);
char *com_get(u16 addr, comtype_t type);
void def_file(char * filename);
u16 hextoi(char * hex);
void disasm_init(u16 addr);
char *disasm(u16 * addr, int flg_dump);

extern sym_t *symbols;
extern com_t *comments;
extern dis_t dis;
#define NEWOPT(a) \
	((memopt[a].w != 0) && (memopt[a].w != dis.curopt.w))

