#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "i8085.h"
#include "dis.h"

u8 mem[0x10000];	/* memory image */

static int flg_dump = 1;
static int flg_default = 0;
static int flg_casette = 0;
static int flg_blkinfo = 0;
static int flg_m100 = 0;
static u16 dis_offset = 0x0000;
static u16 dstart	= 0xFFFF;
static u16 dend	= 0x0000;
static u16 drun	= 0x0000;
static int lcnt	= 0;
static int lmax	= 0x7FFFFFFF;
static char *inp_filename = NULL;
static char *out_filename = NULL;
static char def_filename[160];
static FILE *ifp = NULL;
static FILE *ofp = NULL;

int dis_output(char * src)
{
	while (*src) {
		if (*src == '\n') {
			if (++lcnt >= lmax)
				return 1;
		}
		fputc(*src, ofp);
		src++;
	}
	fputc('\n', ofp);
	if (++lcnt >= lmax)
		return 1;
	return 0;
}

void dis_range(u16 beg, u16 end)
{
	u16 addr = beg;
	u16 old;
	do {
		old = addr;
		if (dis_output(disasm(&addr, flg_dump)))
			return;
	} while (addr >= old && addr <= end);
}

int read_cassette(void)
{
	char name[6+1];
	u16 size;
	u8 csum;
	u8 c;
	int i;

	if ((c = fgetc(ifp)) != 0x55) {
		fprintf(stderr, "file is no cassette (0000:%02X != 55)\n", c);
		return -1;
	}
	while ((c = fgetc(ifp)) == 0x55)
		;

	if (c != 0x7f) {
		fprintf(stderr, "file 1st sync byte missing (%04X:%02X)\n",
			(int)ftell(ifp) - 1, c);
		return -1;
	}

	c = fgetc(ifp);
	switch (c) {
	case 0x9c:
		fprintf(stderr, "file is DOCUMENT image (%04X:%02X)\n",
			(int)ftell(ifp) - 1, c);
		return -1;
	case 0xd3:
		fprintf(stderr, "file is BASIC image (%04X:%02X)\n",
			(int)ftell(ifp) - 1, c);
		return -1;
	case 0xd0:
		break;
	default:
		fprintf(stderr, "file is unknown image type (%04X:%02X)\n",
			(int)ftell(ifp) - 1, c);
		return -1;
	}

	csum = 0;
	for (i = 0; i < 6; i++) {
		c = fgetc(ifp);
		name[i] = c;
		csum += c;
		if (isalnum(c) || c == ' ')
			continue;
		fprintf(stderr, "file has invalid filename (%04X:%02X)\n",
			(int)ftell(ifp) - 1, c);
		return -1;
	}
	name[6] = '\0';

	c = fgetc(ifp);
	csum += c;
	dstart = c;

	c = fgetc(ifp);
	csum += c;
	dstart += 256 * c;

	c = fgetc(ifp);
	csum += c;
	size = c;

	c = fgetc(ifp);
	csum += c;
	size += 256 * c;

	c = fgetc(ifp);
	csum += c;
	drun = c;

	c = fgetc(ifp);
	csum += c;
	drun += 256 * c;

	for (i = 0; i < 4; i++) {
		c = fgetc(ifp);
		csum += c;
		if (c != 0x00) {
			printf("unexpected data in header: (%04X:%02X)\n",
				(int)ftell(ifp) - 1, c);
		}
	}

	csum = -csum;
	c = fgetc(ifp);
	if (c != csum ) {
		fprintf(stderr, "header checksum error: (%04X:%02X != %02X)\n",
			(int)ftell(ifp) - 1, c, csum);
		return -1;
	}

	if (flg_blkinfo) {
		printf("found:%s\n", name);
		printf("top:%u\n", dstart);
		printf("end:%u\n", dstart + size - 1);
		printf("exe:%u\n", drun);
	}

	if ((c = fgetc(ifp)) != 0x00) {
		fprintf(stderr, "file is no cassette (%04X:%02X != 00)\n",
			(int)ftell(ifp) - 1, c);
		return -1;
	}

	while ((c = fgetc(ifp)) == 0x00)
		;

	if (c != 0x55) {
		fprintf(stderr, "file 2nd header missing (%04X:%02X)\n",
			(int)ftell(ifp) - 1, c);
		return -1;
	}

	while ((c = fgetc(ifp)) == 0x55)
		;

	if (c != 0x7f) {
		fprintf(stderr, "file 2nd sync byte missing (%04X:%02X)\n",
			(int)ftell(ifp) - 1, c);
		return -1;
	}

	c = fgetc(ifp);
	if (c != 0x8d) {
		fprintf(stderr, "file data header missing (%04X:%02X)\n",
			(int)ftell(ifp) - 1, c);
		return -1;
	}

	csum = 0;
	dend = dstart;
	while (!feof(ifp) && size-- > 0) {
		c = fgetc(ifp);
		csum += c;
		mem[dend++] = c;
	}
	csum = -csum;
	c = fgetc(ifp);
	if (c != csum) {
		fprintf(stderr, "data checksum error: (%04X:%02X != %02X)\n",
			(int)ftell(ifp) - 1, c, -csum);
		return -1;
	}

	if (0 == drun)
		drun = dstart;

	def_opt(drun, MEM_CODE, 0, OPT_NONE);
	add_sym(drun, SYM_CODE, "EXE");
	return 0;
}

int read_binary(void)
{
	dstart = dis_offset;
	while (!feof(ifp)) {
		int cnt = fread(mem + dis_offset, 1, 0x10000 - dis_offset, ifp);

		if (dstart + cnt - 1 > dend)
			dend = dstart + cnt - 1;

		dis_offset = 0;
	}
	return 0;
}

int main(int ac, char ** av)
{
	int	i;
	char	* dst;

	if (ac < 2) {
		fprintf(stderr, "usage: %s [options] input [output]\n", av[0]);
		fprintf(stderr, "options can be one or more of:\n");
		fprintf(stderr, "\t-a\tASM output (no addr/data columns)\n");
		fprintf(stderr, "\t-c\tfile is in .CAS format\n");
		fprintf(stderr, "\t-d\tinclude default definitions\n");
		fprintf(stderr, "\t-i\tprint block info\n");
		fprintf(stderr, "\t-o xxxx\tload from offset xxxx (hex)\n");
		fprintf(stderr, "\t-jm\tinclude m100.def\n");
		exit(1);
	}

	for (i = 1; i < ac; i++) {
		if (av[i][0] == '-') {
			switch (av[i][1]) {
				case 'a': case 'A':
					flg_dump = 0;
					break;
				case 'c': case 'C':
					flg_casette = 1;
					break;
				case 'd': case 'D':
					flg_default = 1;
					break;
				case 'i': case 'I':
					flg_blkinfo = 1;
					break;
				case 'j': case 'J':
					if (toupper((u8)av[i][2]) == 'M')
						flg_m100  = 1;
					break;
				case 'o': case 'O':
					i++;
					dis_offset = hextoi(av[i]);
					drun = dis_offset;
					break;
				default:
					fprintf(stderr, "unknown option %x\n", av[i][1]);
					exit(1);
					break;
			}
		}
		else
		{
			if (!inp_filename)
				inp_filename = av[i];
			else
				if (!out_filename)
					out_filename = av[i];
		}
	}

	if (!inp_filename) {
		fprintf(stderr, "no input filename given\n");
		exit(1);
	}

	ifp = fopen(inp_filename, "rb");
	if (!ifp) {
		fprintf(stderr, "input file %s not found\n", inp_filename);
		exit(1);
	}
	
	if (out_filename) {
		ofp = fopen(out_filename, "wa");
		if (!ofp) {
			fprintf(stderr, "can't create output file %s\n",
				out_filename);
			exit(1);
		}
	} else {
		ofp = stdout;
	}

	if (flg_casette) {
		if (read_cassette())
			exit(1);
	} else {
		if (read_binary())
			exit(1);
	}

	if (flg_default)
		def_file("8085.def");

	if (flg_m100)	/* include Model 100 symbols? */
		def_file("m100.def");

	strcpy(def_filename, inp_filename);
	dst = strrchr(def_filename, '.');
	if (!dst)
		strcat(def_filename, ".def");
	else
		strcpy(dst, ".def");

	def_file(def_filename);
	dst = strrchr(def_filename, '\\');
	if (!dst)
		dst = strrchr(def_filename, '/');
	if (dst) {
		strcpy(def_filename, dst + 1);
		def_file(def_filename);
	}

	disasm_init(dstart);

	add_sym(drun, SYM_CODE, "ENTRY");

	dis_range(dstart, dend);

	return 0;
}

