#include <stdint.h>
typedef int8_t		i8;	/* offset */
typedef uint8_t		u8;	/* byte */
typedef uint16_t	u16;	/* word */
typedef uint32_t	u32;	/* dword */
typedef int64_t		i64;	/* qword */

typedef struct {	/* flag set */
	unsigned c:1;	/* carry */
	unsigned n:1;	/* negative */
	unsigned p:1;	/* parity/overflow */
	unsigned x:1;	/* unknown */
	unsigned h:1;	/* half carry */
	unsigned y:1;	/* unknown */
	unsigned z:1;	/* zero */
	unsigned s:1;	/* sign */
} flag_t;

typedef union {		/* registers union */
	u8	r8[2];	/* low, high u8 */
	u16	r16;	/* u16 (intel) */
	flag_t	f;	/* flags */
} reg_t;

typedef struct {
	reg_t	bc;
	reg_t	de;
	reg_t	hl;
	reg_t	sp;
	reg_t	af;
	reg_t	pc;
	u8	im;
	u8	IEN;
	u8	data_tmp;
	u8	port_tmp;
	u32	addr_jmp;
	u32	addr_src;
	u32	addr_dst;
	u8	data_xff[4];
} I8085;

#define BC i8085->bc.r16
#define DE i8085->de.r16
#define HL i8085->hl.r16
#define AF i8085->af.r16

#define SP i8085->sp.r16

#define PC i8085->pc.r16

#define B i8085->bc.r8[1]
#define C i8085->bc.r8[0]
#define D i8085->de.r8[1]
#define E i8085->de.r8[0]
#define H i8085->hl.r8[1]
#define L i8085->hl.r8[0]
#define A i8085->af.r8[1]
#define F i8085->af.f
#define I i8085->ri.r8[1]
#define R i8085->ri.r8[0]

#define HS i8085->sp.r8[1]
#define LS i8085->sp.r8[0]

#define fC i8085->af.f.c
#define fN i8085->af.f.n
#define fX i8085->af.f.x
#define fH i8085->af.f.h
#define fY i8085->af.f.y
#define fP i8085->af.f.p
#define fZ i8085->af.f.z
#define fS i8085->af.f.s

extern u8 mem[0x10000];
