'******************************************************************************
'*  io.spin - handle input and output requests sent by pm80
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id: io.spin,v 1.20 2010-03-25 09:40:35 pm Exp $
'******************************************************************************
CON

	io_cmd_out		=	$01		' I/O emulator command output
	io_cmd_in		=	$02		' I/O emulator command input
	io_cmd_break		=	$03		' I/O emulator command break
	io_cmd_kbd		=	$04		' I/O emulator command read keyboard
	io_cmd_get_sid		=	$05		' I/O emulator command get SID status
	io_cmd_set_sod		=	$06		' I/O emulator command set SOD status
	m100_clock		= 	2_400_000	' 2.4MHz system clock
	baud			=	19200		' baud rate

VAR
	long	cog
	long	iocmd
	long	irq
	long	pio
	long	rc
	byte	key
	byte	de[32]					' directory entry
	byte	name[12+1]
	byte	datetime[20+1]
	byte	buff[64]

OBJ
	fdx	: "pcFullDuplexSerial4FC"
	fat	: "fatfs"
	kbd	: "keyboard"
	vid	: "video"
	lcd	: "lcd"
#ifdef TriBladeProp
	xmm	: "TriBladeProp_Blade2_204"		' Extended memory manager
#endif
#ifdef RamBladeProp
	xmm	: "RamBlade_203_002"			' Extended memory manager
#endif
#ifdef DracBladeProp
	xmm	: "Dracblade"				' Extended memory manager
#endif

PUB start(pins) : okay | i, dent, sec
	stop

	cmd_ptr := @iocmd
	pio_ptr := @pio
	irq_ptr := @irq
	iobuff_ptr := @iobuff
	rtc_1990_ptr := @rtc_1990_data

	' ***************************************************************
	' Start the FAT driver and the SD SPI object (must be first)
	'
	' cog #1
	'
	okay := \fat.start(byte[pins][4], byte[pins][5], byte[pins][6], byte[pins][7])
	if okay < 0
		return -1
	spi_ptr := fat.get_SPI_command
	iobuff_ptr := fat.get_memory

	' ***************************************************************
	' Start the full duplex serial object
	'
	' cog #2
	'
	fdx.AddPort(0, byte[pins][0], byte[pins][1], -1, -1, 0, 0, baud)
	if byte[pins][2] <> $ff
		fdx.AddPort(1, byte[pins][2], byte[pins][3], -1, -1, 0, 0, baud)
	okay := fdx.Start
	if okay < 0
		return
	tx_head := fdx.get_tx_head
	tx_tail := fdx.get_tx_tail
	tx_buffer := fdx.get_tx_buffer
	rx_head := fdx.get_rx_head
	rx_tail := fdx.get_rx_tail
	rx_buffer := fdx.get_rx_buffer
	fdx.str(0,string("Tandy M100 emulator on Parallax Propeller.",13,10))

	' ***************************************************************
	' Start the m100 keyboard driver
	'
	' cog #3
	'
	okay := kbd.startx(byte[pins][8], byte[pins][9], %0_000_100, 40, @irq, @rtc_1990_data)
	if okay < 0
		return -1
	kbd_states_ptr := kbd.states_ptr
	fdx.str(0, string("KEYBOARD cog setup.",13,10))

	' ***************************************************************
	' Start the *Blade XMM handler object
	'
	' cog #4
	'
	okay := \xmm.start
	if okay < 0
		return -1
	fdx.str(0, string("XMM cog setup.",13,10))

	' ***************************************************************
	' mount the FAT, i.e. check the format of the SD media
	'
	okay := \fat.mount
	if okay < 0
		return -1
	fdx.str(0, string("FAT mounted.",13,10))

	' *********************************************************
	' find the M100.ROM file and load it into the ROM region
	'
	dent := open_file(string("M100.ROM"))
	if dent < 0
		return -1
	bytemove(@de, dent, 32)
	sec := fat.get_dent_first_sector(@de)
	fdx.str(0, string("Found "))
	fdx.str(0, @name)
	fdx.str(0, string(13,10))
	repeat i from $0000 to $7e00 step $200
		fat.read_sector(sec)
		xmm.DoCmd("W", iobuff_ptr, i, $200)
		sec := fat.get_dent_next_sector(@de, sec)
	longfill(iobuff_ptr, 0, $200/4)
	repeat i from $8000 to $fe00 step $200
		xmm.DoCmd("W", iobuff_ptr, i, $200)

	byte[iobuff_ptr] := "2"						' Y2K patch
	xmm.DoCmd("W", iobuff_ptr, $5a53, 1)
	byte[iobuff_ptr] := "0"
	xmm.DoCmd("W", iobuff_ptr, $5a56, 1)

	byte[iobuff_ptr] := 0
	xmm.DoCmd("W", iobuff_ptr, $7d39, 1)

	' *********************************************************
	' find an M100.RAM file and remember its first sector
	'
	dent := open_file(string("M100.RAM"))
	if dent > 0
		bytemove(@de, dent, 32)
		ram_sec := fat.get_dent_first_sector(@de)
		fdx.str(0, string("Found "))
		fdx.str(0, @name)
		fdx.str(0, string(13,10))

	' ***************************************************************
	' Start the I/O handler cog
	'
	' cog #5
	'
	okay := cog := COGNEW(@entry, pins) + 1
	if okay =< 0
		return -1
	fdx.str(0, string("I/O cog setup.",13,10))

	' ***************************************************************
	' Start the LCD i/o port handler and emulation
	'
	' cog #6
	okay := lcd.start(@pio, @iocmd)
	if okay =< 0
		return -1
	fdx.str(0, string("LCD cog setup.",13,10))

	' ***************************************************************
	' Start the m100 video driver
	'
	' cog #7
	'
	okay := vid.start(lcd.get_bitmap, byte[pins][11])
	if okay =< 0
		return -1
	fdx.str(0, string("VIDEO cog setup.",13,10))

	' Stop the XMM handler
	\xmm.stop

	return 0

PUB stop
'' Stop CPU simulator - frees a cog
	if cog
		cogstop(cog~ - 1)

PRI open_file(fname_ptr) : dent | size, sec
	dent := fat.find_file(fname_ptr)
	if dent < 0
		return -1
	bytemove(@de, dent, 32)
	fat.get_dent_name(dent, @name)
	size := fat.get_dent_size(@de)
	sec := fat.get_dent_first_sector(@de)

PUB chr(ch)
	fdx.tx(0,ch)

PUB str(strp)
	fdx.str(0,strp)

PUB dec(value) | _i
'' Print a decimal number
	result := 0
	if value < 0
		chr("-")
		if (value == NEGX)
			chr("2")
			value += 2_000_000_000
		-value
	_i := 1_000_000_000
	repeat 10
		if value => _i
			chr(value / _i + "0")
			value //= _i
			result~~
		elseif result or _i == 1
			chr("0")
		_i /= 10

PUB command_addr
'' Return the address of the command long
	return @iocmd

PUB irq_addr
'' Return the address of the irq long
	return @irq

DAT
		org	0
entry
		mov	t1, PAR
		add	t1, #10				' point to auPin
		rdbyte	au_pin, t1			' get the value
		mov	au_mask, #1
		shl	au_mask, au_pin
		mov	DIRA, au_mask
		or	OUTA, au_mask

		mov	data, #0
		wrlong	data, cmd_ptr

command		mov	t1, pio_timer		WZ
	if_z	jmp	#:skip
		mov	t1, CNT
		sub	t1, pio_cnt0			' timer start time
		cmps	t1, pio_timer		WZ, WC	' reached terminal count?
 	if_ae	call	#pio_timer_cb			' call the timer callback

:skip		rdlong	data, cmd_ptr
		mov	cmd, data
		shr	cmd, #24		WZ
	if_z	jmp	#command

		mov	port, data
		shr	port, #8			' shift port # to bits 7..0
		and	port, #$ff			' mask bits 7..0
		add	cmd, cmd_table_ptr
		rdbyte	cmd, cmd			' dispatch on command
		jmp	cmd
return_ff
		mov	data, #$ff			' unmapped port result
return_data
		and	data, #$ff			' mask bits 7..0
		or	cmd, data			' put data into bits 7..0
command_0
		and	cmd, low_24bits
		wrlong	cmd, cmd_ptr
ignore_cmd
		call	#uart_stream			' poll the RX port and stuff the UART
		jmp	#command

'*******************************************
'
' out port, data
'
command_out
		mov	cmd, data
		and	data, #$ff			' mask output byte
		mov	t1, wr_port_ptr			' dispatch on write port
		add	t1, port
		rdbyte	t1, t1
		jmp	t1

'*******************************************
'
' inp port
'
command_inp
		mov	cmd, data
		andn	cmd, #$ff			' clear result byte
		mov	t1, rd_port_ptr			' dispatch on read port
		add	t1, port
		rdbyte	t1, t1
		jmp	t1

'*******************************************
'
' Break - dump registers
'
command_break
		mov	cmd, data
		call	#break
		jmp	#command_0

'*******************************************
'
' Get current SID line status
'
command_get_sid
		mov	cmd, data
		andn	cmd, #$ff		' clear return value
		jmp	#command_0

'*******************************************
'
' Set current SOD line status
'
command_set_sod
		mov	cmd, data
		jmp	#command_0


'*******************************************
' 82: Option register read
' I don't know much about the option register.
' All that is known that it usually
' (or always?) returns the value $A2
'
m100_option_r
		mov	data, #$a2
		jmp	#return_data

'*******************************************
' a0/a8: Read the modem status register
'
' bit   meaning
' ------------------------
'  0    ???
'  1    ???
'  2    ???
'  3    ???
'  4    ???
'  5    ???
'  6    ???
'  7    ???
'
modem_status_r
		mov	data, #$a0
		jmp	#return_data

'*******************************************
' a0/a8: Write the modem control register
'
' bit   meaning
' ------------------------
'  0    ???
'  1    ???
'  2    ???
'  3    ???
'  4    ???
'  5    ???
'  6    ???
'  7    ???
'
modem_control_w
		jmp	#command_0

'*******************************************
' b0/b8: Read PIO status register
'
' bit   meaning
' ------------------------
'  0    port A interrupt request
'  1    port A buffer full/empty (input/output)
'  2    port A interrupt enabled
'  3    port B interrupt request
'  4    port B buffer full/empty (input/output)
'  5    port B interrupt enabled
'  6    Timer interrupt (status of TC pin)
'  7    unused
'
pio_status_r
		mov	data, pio_status
		or	data, #%10000000
		jmp	#return_data

'*******************************************
' b0/b8: Write PIO 8155 control register
' bit 0   - Direction of port A (0 input, 1 output)
' bit 1   - Direction of port B (0 input, 1 output)
' bit 2,3 - Port C mode
'           00 = port C all input
'           01 = port C alt 3
'           10 = port C alt 4
'           11 = port C all output
' bit 4   - Enable port A interrupt (1 enable)
' bit 5   - Enable port B interrupt (1 enable)
' bit 6,7 - Timer mode
'           00 = NOP:   Does not affect counter operations
'           01 = STOP:  Stops the timer if it is running
'                       NOP if the timer is not running
'           10 = TC:    After TC: Stops timer when it reaches TC
'                       NOP if the timer is not running
'           11 = START: If the timer is not running, loads the mode
'                       and the count length, and immediately starts
'                       timer operation.
'                       If the timer is running, loads a new mode
'                       and the count length, and starts timer
'                       operation immediately after TC is reached.
pio_control_w
		test	data, #%01000000	WZ
		test	data, #%10000000	WC
 if_z_and_nc	jmp	#:mode0
 if_nz_and_nc	jmp	#:mode1
 if_z_and_c	jmp	#:mode2

:mode3		' START If the timer is not running, loads the mode
		'       and the count length, and immediately starts
		'       timer operation.
		'       If the timer is running, loads a new mode
		'       and the count length, and starts timer
		'       operation immediately after TC is reached.
		mov	pio_cnt0, CNT			' get current CNT as timer start time
		mov	x, i2_400_000			' get M100 clock
		mov	y, pio_reload
		shr	y, #1			WC
		addx	y, #0
		call	#div32				' divide
		rdlong	x, #0				' get prop clkfreq
		call	#div32				' divide by result
		mov	pio_timer, y			' timer rate in Prop cycles
		jmp	#:exit

:mode2		' TC    After TC: Stops thimer when it reaches TC
		'       NOP if the timer is not running
		cmp	pio_timer, #0		WZ
	if_nz	jmp	#:mode3
		jmp	#:exit

:mode1		' STOP  Stops the timer if it is running
		'       NOP if the timer is not running
		mov	pio_timer, #0			' stop timer
:mode0		' NOP   Does not affect counter operations
:exit		andn	pio_status, #%01000000		' reset the TC flag bit
		mov	pio_control, data
		jmp	#command_0

'*******************************************
' Read PIO port A
'
pio_port_a_r
		rdbyte	data, pio_ptr
		jmp	#return_data

'*******************************************
' Write PIO port A
'
' Port A holds 8 bit data for printer output,
' keyboard column strobe, and LCD driver selection.
' In addition, the first 5 bits of this port are used to
' control the 1990 real time clock chip.
'
' The configuration of the 5 bits is:
' bit 0   - C0
' bit 1   - C1
' bit 2   - C2
' bit 3   - Clock 1990
' bit 4   - Serial data into clock chip
' bit 5   - unused
' bit 6   - unused
' bit 7   - unused
pio_port_a_w
		rdbyte	t1, pio_ptr
		xor	t1, data
		test	t1, #%00001000		WZ	' bit 3 changed?
	if_z	jmp	#:exit
		test	data, #%00001000	WZ	' rising edge of RTC 1990 clock?
	if_z	jmp	#:exit
		rol	rtc_1990_bit, #1	WC	' next bit to test
	if_c	mov	rtc_1990_idx, #4		' 2nd long containing the remaining bits
		mov	t2, rtc_1990_idx		' get index
		add	t2, rtc_1990_ptr
		rdlong	t2, t2
		test	t2, rtc_1990_bit	WZ
		rdlong	pio_port, pio_ptr
		muxnz	pio_port, port_c_0		' set or clear port C bit 0
		wrlong	pio_port, pio_ptr
:exit
		wrbyte	data, pio_ptr			' set new value for port A
		jmp	#command_0


'*******************************************
' Read PIO port B
'
pio_port_b_r
		mov	ptr, pio_ptr
		add	ptr, #1
		rdbyte	data, ptr
		jmp	#return_data

'*******************************************
' Write PIO port B
' bit 0 - column 9 select for the keyboard, and also
'       - CS-28 line of the LCD
' bit 1 - CS-29 line of the LCD
' bit 2 - beep toggle (1 = data from bit 5, 0 = data from 8155 timer)
' bit 3 - serial toggle (1 = modem, 0 = RS232)
' bit 4 - software on/off switch for computer
' bit 5 - data to beeper if bit 2 is set
' bit 6 - ~DTR for RS232
' bit 7 - ~RTS for RS232
'
pio_port_b_w
		mov	ptr, pio_ptr
		add	ptr, #1
		rdbyte	t1, ptr
		xor	t1, data
		test	t1, #%00100100		WZ	' changed beeper?
	if_z	jmp	#:exit
		test	data, #%00000100	WC	' data from timer?
	if_nc	test	pio_status, #%01000000	WZ	' check PIO status bit 6
	if_c	test	data, #%00100000	WZ	' check data bit 5
		muxnz	OUTA, au_mask
:exit		wrbyte	data, ptr
		jmp	#command_0

'*******************************************
' Read PIO port C
'
' bit 0   - serial data input from clock chip 1990
' bit 1   - ~BUSY signal from printer
' bit 2   - BUSY signal from printer
' bit 3   - data from BCR (bar code reader)
' bit 4   - ~CTS line from RS232
' bit 5   - ~DSR line from RS232
' bit 6,7 - not available on 8155
'
pio_port_c_r
		mov	ptr, pio_ptr
		add	ptr, #2
		rdbyte	data, ptr
		test	data, #%00000010	WZ	' make sure the busy/~busy signals are the opposite
		muxz	data, #%00000100
		jmp	#return_data

'*******************************************
' Write PIO port C
'
pio_port_c_w
		mov	ptr, pio_ptr
		add	ptr, #2
		mov	t1, data
		and	t1, #%00111111
		wrbyte	t1, ptr
		jmp	#command_0


'*******************************************
' PIO timer LSB read
'
pio_timer_l_r
		cmp	pio_timer, #0		WZ
	if_z	jmp	#:exit
		' TODO: calculate elapsed time and adjust
		' pio_counter to 2 * cycles + (pio_reload & 1)
		mov	pio_count, pio_timer
		test	pio_reload, #1		WC
		addx	pio_count, #0
:exit		mov	data, pio_count
		jmp	#return_data


'*******************************************
' PIO timer LSB write
'
pio_timer_l_w
		andn	pio_reload, #$ff
		or	pio_reload, data
		' TODO: calculate speaker clock and change it
		jmp	#command_0

'*******************************************
' PIO timer MSB read
'
pio_timer_h_r
		cmp	pio_timer, #0		WZ
	if_z	jmp	#:exit
		' TODO: calculate elapsed time and adjust
		' pio_counter to 2 * cycles + (pio_reload & 1)
		mov	pio_count, pio_timer
		test	pio_reload, #1		WC
		addx	pio_count, #0
:exit		mov	data, pio_count
		shr	data, #8
		and	data, #$3f
		or	data, pio_tmrctl
		jmp	#return_data

'*******************************************
' PIO timer MSB write
'
pio_timer_h_w
		mov	pio_tmrctl, data
		and	pio_tmrctl, #%11000000
		ror	pio_reload, #8
		movs	pio_reload, data
		xor	pio_reload, pio_tmrctl
		rol	pio_reload, #8
		mov	pio_tc, #1
		jmp	#command_0

'*******************************************
' UART data read
'
uart_data_r
		andn	uart_status, #%00100000		' clear input buffer full status
		rdlong	t1, irq_ptr
		andn	t1, #%00100000			' clear request of RST 6.5
		mov	data, uart_data
		wrlong	t1, irq_ptr
		jmp	#return_data

'*******************************************
' UART data write
'
uart_data_w
		cmp	data, #$11		WZ	' XON?
	if_z	mov	uart_xoff, #0
	if_z	jmp	#:exit
		cmp	data, #$13		WZ
	if_z	mov	uart_xoff, #1
	if_z	jmp	#:exit
		mov	char, data
		call	#tx
:exit		jmp	#command_0

'*******************************************
' UART status register
' bit 0 - data on telephone line (carrier detect)
' bit 1 - overrun error
' bit 2 - framing error
' bit 3 - parity error
' bit 4 - transmitter buffer empty
' bit 5 - ring line on modem connector
' bit 6 - unused
' bit 7 - low power signal from power supply (0 = low power)
'
uart_status_r
		mov	data, uart_status		' get status
		or	data, #%00010000		' transmitter buffer is always empty
		or	data, #%10000000		' always good power
		jmp	#return_data

'*******************************************
' UART control register
' bit 0    - number of stop bits (0 = 2, 1 = 1.5)
' bit 1    - parity (1 even, 0 odd)
' bit 2    - parity enable (1 no parity, 0 parity enabled)
' bit 3,4  - data length
'            00 - 5 bits
'            01 - 6 bits
'            10 - 7 bits
'            11 - 8 bits
' bit 5-7  - unused?
'
uart_control_w
		mov	uart_control, data
		jmp	#command_0

'*******************************************
' Keyboard status
'
kbd_status_r
		cmp	port, #$e9		WC	' keyboard scan?
	if_c	mov	data, #$ff
	if_nc	mov	data, #$00
		mov	t4, kbd_states_ptr
		rdlong	pio_port, pio_ptr
		test	pio_port, #%1_00000000	WZ	' key row 8 to test?
	if_z	add	t4, #8
	if_z	rdbyte	t1, t4
	if_z	andn	data, t1
	if_z	jmp	#return_data
		mov	t2, #1				' bit 0
		mov	t3, #8				' 8 rows to scan
:loop		test	pio_port, t2		WZ	' port A/B bit set?
	if_z	rdbyte	t1, t4
	if_z	andn	data, t1			' reset bits
		add	t4, #1				' next keyboard state
		rol	t2, #1				' next bit
		djnz	t3, #:loop
:exit
		jmp	#return_data

'*******************************************
' Miscellaneous device select
' bit 0    - ROM select (0 standard ROM, 1 option ROM)
' bit 1    - strobe signal to printer (1 strobe low, 0 strobe high)
' bit 2    - strobe for clock chip 1990
' bit 3    - remote plug control signal
' bit 4-7  - unused?
'
misc_control_w
		mov	t1, misc_control
		xor	t1, data		WZ
	if_z	jmp	#:exit
		test	t1, #%0001		WZ
	if_z	jmp	#:no_ext_rom
		' TODO: select BASIC or EXTENSION ROM
:no_ext_rom	test	t1, #%0010		WZ
	if_z	jmp	#:no_prt_strobe
		' TODO: strobe printer, i.e. write output
:no_prt_strobe	test	t1, #%0100		WZ
	if_z	jmp	#:no_rtc_strobe
		test	data, #%0100		WZ	' rising edge of RTC 1990 strobe?
	if_z	jmp	#:no_rtc_strobe
		rdlong	pio_port, pio_ptr
		mov	t2, pio_port			' get pio port A
		and	t2, #%111		WZ	' bits 2..0
	if_z	jmp	#:rtc_mode_c0
		cmp	t2, #%001		WZ
	if_z	jmp	#:rtc_mode_c1
		cmp	t2, #%010		WZ
	if_nz	cmp	t2, #%011		WZ
	if_z	jmp	#:rtc_mode_c23
		cmp	t2, #%101		WZ
	if_z	jmp	#:rtc_mode_c5
		mov	rtc_1990_idx, #4		' end of shift register file
		mov	rtc_1990_bit, #1
		jmp	#:no_rtc_strobe
:rtc_mode_c5						' WRITE?
		' TODO: is this a write mode?
		jmp	#:no_rtc_strobe
:rtc_mode_c23						' READ!?
		mov	rtc_1990_idx, #0		' reset shift register
		mov	rtc_1990_bit, #1
		jmp	#:no_rtc_strobe
:rtc_mode_c1						' SHIFT
		mov	rtc_1990_idx, #0		' reset shift register
		mov	rtc_1990_bit, #1
		mov	t2, rtc_1990_ptr
		rdlong	t2, t2
		test	t2, rtc_1990_bit	WZ	' test bit 0 of the shift register
		muxnz	pio_port, port_c_0		' set/clear port C bit 0
		wrlong	pio_port, pio_ptr
:rtc_mode_c0						' NOP
:no_rtc_strobe
		test	t1, #%1000		WZ	' remote plug changed?
	if_z	jmp	#:exit
		' TODO: turn remote plug on/off
:exit		mov	misc_control, data
		jmp	#command_0

'*******************************************
' Write to the M100.RAM file
'
ram_w
		cmp	ram_sec, #0		WZ	' do we have a start sector?
	if_z	jmp	#command_0			' ignore the write
		mov	t1, ram_offs
		add	t1, iobuff_ptr
		wrbyte	data, t1
		add	ram_offs, #1
		test	ram_offs, #$1ff		WZ	' 512 bytes written?
	if_nz	jmp	#command_0			' not yet
		call	#write_iobuff
		mov	ram_offs, #0
		add	ram_sec, #1
		add	ram_size, #1
		jmp	#command_0

'*******************************************
' UART callback: receive character and
' set the interrupt RST 6.5
'
uart_stream
		test	uart_status, #%00100000	WZ	' input buffer full?
	if_nz	jmp	#uart_stream_ret
		rdlong	fdx_head, rx_head
		rdlong	fdx_tail, rx_tail
		cmp	fdx_head, fdx_tail	WZ	' head == tail?
	if_z	jmp	#uart_stream_ret
		mov	fdx_temp, fdx_tail
		add	fdx_temp, rx_buffer
		rdbyte	uart_data, fdx_temp		' get character
		add	fdx_tail, #1
		and	fdx_tail, #63			' wrap tail pointer
		wrlong	fdx_tail, rx_tail
		or	uart_status, #%00100000		' set input buffer full status
		rdlong	t1, irq_ptr
		or	t1, #%00100000			' request RST 6.5
		wrlong	t1, irq_ptr
uart_stream_ret
		ret

'*******************************************
'
' Timer callback
'
pio_timer_cb
		add	pio_cnt0, pio_timer		' reload timer start time
		rdlong	pio_port, pio_ptr		' get PIO ports
		test	pio_tmrctl, #%01000000	WZ
		test	pio_tmrctl, #%10000000	WC
 if_z_and_nc	jmp	#:mode0
 if_nz_and_nc	jmp	#:mode1
 if_z_and_c	jmp	#:mode2

:mode3		' Outputs a pulse each time the preset TC value is reached.
		' Automatically loads the programmed count lengths, and restarts from the beginning.
:reload		mov	x, i2_400_000			' get M100 clock
		mov	y, pio_reload			' divide by (reload/2 + reload&1)
		shr	y, #1			WC
		cmp	pio_tc, #0		WZ
	if_z	addx	y, #0
		call	#div32				' divide
		rdlong	x, #0				' get prop clkfreq
		call	#div32				' divide by result
		mov	pio_timer, y			' CNT delta until next event
		jmp	#:exit

:mode2		' Outputs a pulse when the TC value is reached.
		test	pio_status, #%01000000	WZ
	if_z	mov	pio_tc, #1
	if_z	jmp	#:reload
		mov	pio_timer, #0			' shut the timer
		jmp	#:exit

:mode1		' Outputs a low level signal in the latter half of a count period.
		' Automatically loads the programmed count length, and restarts
		' counting when the TC value is reached.
		xor	pio_tc, #1
		jmp	#:reload

:mode0		' Outputs a low level signal in the latter half (Note 1) of a count period.
		mov	pio_tc, #0
:exit
		cmp	pio_tc, #0		WZ
		muxnz	pio_status, #%01000000		' set or clear terminal count flag
		test	pio_port, port_b_2	WC	' test bit 2 of port B (timer to audio)
	if_nc	muxnz	OUTA, au_mask			' bit 2 is set: output TC to audio
pio_timer_cb_ret
		ret

'*******************************************
' Divide x by y
' quotient y
' remainder x
'
' clobbers: t1 and t2
'
div32
		mov	y, y		WZ		' division by zero?
	if_z	mov	y, x				' return x
	if_z	jmp	#div32_ret
		mov	t1, #0
		mov	t2, #32				' 32 quotient bits
:loop1		shr	y, #1		WZ, WC
		rcr	t1, #1
	if_nz	djnz	t2, #:loop1
:loop2		cmpsub	x, t1		WC
		rcl	y, #1
		shr	t1, #1
		djnz	t2, #:loop2
div32_ret
		ret

'*******************************************
' Write the contents of iobuff_ptr to
' the SD card's M100.RAM file
'
write_iobuff
		mov	ptr, spi_ptr			' get pointer to SPI block
		add	ptr, #4				' point to block number
		wrlong	ram_sec, ptr
		add	ptr, #4				' point to hub RAM address
		wrlong	iobuff_ptr, ptr
		sub	ptr, #2*4			' point to SPI command
		mov	t1, #"w"			' issue write command
		wrlong	t1, ptr
:wait		rdlong	t1, ptr			WZ
	if_nz	jmp	#:wait
		mov	hexval, ram_sec
		call	#conout_hex_l
		call	#conout_crlf
write_iobuff_ret
		ret

'*******************************************
' Send a character to the first UART
'
tx
		rdlong	fdx_head, tx_head
		mov	fdx_temp, fdx_head
		add	fdx_head, #1
		rdlong	fdx_tail, tx_tail
		and	fdx_head, #15
		cmp	fdx_head, fdx_tail	WZ
	if_z	jmp	#tx				' wait until there's room in the buffer
		add	fdx_temp, tx_buffer
		wrbyte	char, fdx_temp
		wrlong	fdx_head, tx_head		' increment the tx_head
tx_ret
		ret

'*******************************************
'
' Output a byte in hexval to the console
'
conout_hex_b
		mov	hexcnt, #2
		shl	hexval, #24
		jmp	#conout_hex

'*******************************************
'
' Output a word in hexval to the console
'
conout_hex_w
		mov	hexcnt, #4
		shl	hexval, #16
		jmp	#conout_hex

'*******************************************
'
' Output a long in hexval to the console
'
conout_hex_l
		mov	hexcnt, #8
conout_hex
		mov	char, hexval
		shr	char, #28
		cmp	char, #10		WZ, WC
	if_b	add	char, #48
	if_ae	add	char, #55
		call	#tx
		shl	hexval, #4
		djnz	hexcnt, #conout_hex
conout_hex_ret
conout_hex_l_ret
conout_hex_w_ret
conout_hex_b_ret
		ret

'*******************************************
'
' Output a string at strptr to the console
'
conout_string
		rdbyte	char, strptr	WZ
		add	strptr, #1
	if_z	jmp	#conout_string_ret
		call	#tx
		jmp	#conout_string
conout_string_ret
		ret

'*******************************************
'
' Print a CR and LF to the console
'
conout_crlf
		mov	char, #13
		call	#tx
		mov	char, #10
		call	#tx
conout_crlf_ret
		ret

'*******************************************
'
' Print a space to the console
'
conout_space
		mov	char, #32
		call	#tx
conout_space_ret
		ret

'*******************************************
' Break - dump registers
'
break
		mov	t2, data
		shr	t2, #8				' extract "port"
		and	t2, low_16bits			' 15..0 = address of the register file
		mov	strptr, msg_break_ptr
		mov	t1, #4				' 4 register pairs
:loop		call	#conout_string
		rdword	hexval, t2
		add	t2, #2
		call	#conout_hex_w
		djnz	t1, #:loop

		call	#conout_string
		rdbyte	hexval, t2
		add	t2, #1
		call	#conout_hex_b

		call	#conout_string
		rdbyte	hexval, t2
		add	t2, #1
		call	#conout_hex_b

		call	#conout_string
		rdword	hexval, t2
		add	t2, #2
		call	#conout_hex_w

		call	#conout_string
		rdword	hexval, t2
		add	t2, #2
		call	#conout_hex_w

		call	#conout_crlf
break_ret
		ret

'*******************************************
' Initialized data
'
pio_port	long	0				' the long read from pio_ptr
pio_status	long	0
pio_control	long	0
pio_timer	long	0
pio_tmrctl	long	0
pio_tc		long	0
pio_cnt0	long	0
pio_reload	long	0
pio_count	long	0

uart_status	long	0
uart_data	long	0
uart_control	long	0
uart_xoff	long	0

misc_control	long	0

rtc_1990_idx	long	0				' index into shift register
rtc_1990_bit	long	1				' bitmask for shift register

ram_offs	long	0				' offset into iobuff_ptr
ram_size	long	0				' M100.RAM sectors written so far

kbd_states_ptr	long	0-0
cmd_ptr		long	0-0
iobuff_ptr	long	0-0
irq_ptr		long	0-0
pio_ptr		long	0-0
rtc_1990_ptr	long	0-0
au_pin		long	0-0
ram_sec		long	0-0
spi_ptr		long	0-0
' pcFullDuplexSerial4FC
tx_head		long	0-0
tx_tail		long	0-0
tx_buffer	long	0-0
rx_head		long	0-0
rx_tail		long	0-0
rx_buffer	long	0-0
wr_port_ptr	long	@@@wr_port
rd_port_ptr	long	@@@rd_port
cmd_table_ptr	long	@@@cmd_table
msg_break_ptr	long	@@@msg_break

low_24bits	long	$00ffffff
low_16bits	long	$0000ffff
port_c_0	long	1<<16
port_b_2	long	1<<10
i2_400_000	long	2_400_000			' M100 clock frequency

'*******************************************
' Uninitialized data
'
cmd		res	1
port		res	1
data		res	1
ptr		res	1
au_mask		res	1

t1		res	1
t2		res	1
t3		res	1
t4		res	1
x		res	1
y		res	1

fdx_head	res	1
fdx_tail	res	1
fdx_temp	res	1

char		res	1
strptr		res	1
hexval		res	1
hexcnt		res	1

		fit	$1f0
rtc_1990_data	byte	$50,$59,$23,$31,$c1,0,0,0		' 40 bits shift register
iobuff		long	0[512*2/4]

wr_port
		byte	command_0,	command_0,	command_0,	command_0	' 00
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0

		byte	command_0,	command_0,	command_0,	command_0	' 10
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0

		byte	command_0,	command_0,	command_0,	command_0	' 20
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0

		byte	command_0,	command_0,	command_0,	command_0	' 30
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0

		byte	ram_w,		command_0,	command_0,	command_0	' 40
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0

		byte	command_0,	command_0,	command_0,	command_0	' 50
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0

		byte	command_0,	command_0,	command_0,	command_0	' 60
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0

		byte	command_0,	command_0,	command_0,	command_0	' 70
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0

		byte	command_0,	command_0,	command_0,	command_0	' 80
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0

		byte	command_0,	command_0,	command_0,	command_0	' 90
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0

		byte	modem_control_w,command_0,	command_0,	command_0	' a0
		byte	command_0,	command_0,	command_0,	command_0
		byte	modem_control_w,command_0,	command_0,	command_0
		byte	command_0,	command_0,	command_0,	command_0

		byte	pio_control_w,	pio_port_a_w,	pio_port_b_w,	pio_port_c_w	' b0
		byte	pio_timer_l_w,	pio_timer_h_w,	command_0,	command_0
		byte	pio_control_w,	pio_port_a_w,	pio_port_b_w,	pio_port_c_w
		byte	pio_timer_l_w,	pio_timer_h_w,	command_0,	command_0

		byte	uart_data_w,	uart_data_w,	uart_data_w,	uart_data_w	' c0
		byte	uart_data_w,	uart_data_w,	uart_data_w,	uart_data_w
		byte	uart_data_w,	uart_data_w,	uart_data_w,	uart_data_w
		byte	uart_data_w,	uart_data_w,	uart_data_w,	uart_data_w

		byte	uart_control_w,	uart_control_w,	uart_control_w,	uart_control_w	' d0
		byte	uart_control_w,	uart_control_w,	uart_control_w,	uart_control_w
		byte	uart_control_w,	uart_control_w,	uart_control_w,	uart_control_w
		byte	uart_control_w,	uart_control_w,	uart_control_w,	uart_control_w

		byte	misc_control_w,	misc_control_w,	misc_control_w,	misc_control_w	' e0
		byte	misc_control_w,	misc_control_w,	misc_control_w,	misc_control_w
		byte	misc_control_w,	misc_control_w,	misc_control_w,	misc_control_w
		byte	misc_control_w,	misc_control_w,	misc_control_w,	misc_control_w

		byte	ignore_cmd,	ignore_cmd,	ignore_cmd,	ignore_cmd	' f0
		byte	ignore_cmd,	ignore_cmd,	ignore_cmd,	ignore_cmd
		byte	ignore_cmd,	ignore_cmd,	ignore_cmd,	ignore_cmd
		byte	ignore_cmd,	ignore_cmd,	ignore_cmd,	ignore_cmd

rd_port
		byte	return_ff,	return_ff,	return_ff,	return_ff	' 00
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff

		byte	return_ff,	return_ff,	return_ff,	return_ff	' 10
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff

		byte	return_ff,	return_ff,	return_ff,	return_ff	' 20
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff

		byte	return_ff,	return_ff,	return_ff,	return_ff	' 30
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff

		byte	return_ff,	return_ff,	return_ff,	return_ff	' 40
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff

		byte	return_ff,	return_ff,	return_ff,	return_ff	' 50
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff

		byte	return_ff,	return_ff,	return_ff,	return_ff	' 60
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff

		byte	return_ff,	return_ff,	return_ff,	return_ff	' 70
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff

		byte	return_ff,	return_ff,	m100_option_r,	return_ff	' 80
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff

		byte	return_ff,	return_ff,	return_ff,	return_ff	' 90
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff

		byte	modem_status_r,	return_ff,	return_ff,	return_ff	' a0
		byte	return_ff,	return_ff,	return_ff,	return_ff
		byte	modem_status_r,	return_ff,	return_ff,	return_ff
		byte	return_ff,	return_ff,	return_ff,	return_ff

		byte	pio_status_r,	pio_port_a_r,	pio_port_b_r,	pio_port_c_r	' b0
		byte	pio_timer_l_r,	pio_timer_h_r,	return_ff,	return_ff
		byte	pio_status_r,	pio_port_a_r,	pio_port_b_r,	pio_port_c_r
		byte	pio_timer_l_r,	pio_timer_h_r,	return_ff,	return_ff

		byte	uart_data_r,	uart_data_r,	uart_data_r,	uart_data_r	' c0
		byte	uart_data_r,	uart_data_r,	uart_data_r,	uart_data_r
		byte	uart_data_r,	uart_data_r,	uart_data_r,	uart_data_r
		byte	uart_data_r,	uart_data_r,	uart_data_r,	uart_data_r

		byte	uart_status_r,	uart_status_r,	uart_status_r,	uart_status_r	' d0
		byte	uart_status_r,	uart_status_r,	uart_status_r,	uart_status_r
		byte	uart_status_r,	uart_status_r,	uart_status_r,	uart_status_r
		byte	uart_status_r,	uart_status_r,	uart_status_r,	uart_status_r

		byte	kbd_status_r,	kbd_status_r,	kbd_status_r,	kbd_status_r	' e0
		byte	kbd_status_r,	kbd_status_r,	kbd_status_r,	kbd_status_r
		byte	kbd_status_r,	kbd_status_r,	kbd_status_r,	kbd_status_r
		byte	kbd_status_r,	kbd_status_r,	kbd_status_r,	kbd_status_r

		byte	ignore_cmd,	ignore_cmd,	ignore_cmd,	ignore_cmd	' f0
		byte	ignore_cmd,	ignore_cmd,	ignore_cmd,	ignore_cmd
		byte	ignore_cmd,	ignore_cmd,	ignore_cmd,	ignore_cmd
		byte	ignore_cmd,	ignore_cmd,	ignore_cmd,	ignore_cmd

cmd_table
		byte	command_0		' 0
		byte	command_out		' io_cmd_out
		byte	command_inp		' io_cmd_in
		byte	command_break		' io_cmd_break
		byte	command_0		' io_cmd_kbd
		byte	command_get_sid		' io_cmd_get_sid
		byte	command_set_sod		' io_cmd_set_sod

msg_break	byte	"BC:",0
		byte	" DE:",0
		byte	" HL:",0
		byte	" AF:",0
		byte	" IM:",0
		byte	" IM'",0
		byte	" SP:",0
		byte	" PC:",0

''      MIT LICENSE
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}
