'******************************************************************************
'*  io2.spin - Handle the LC display output
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id$
'******************************************************************************
CON

io_cmd_out		=	$01
io_cmd_in		=	$02
LCD_DRIVER_W		=	50

VAR
	long	cog
	long	params[4]

PUB start(pio, iocmd) : okay
	stop
	params[0] := @bitmap
	params[1] := @lcd
	params[2] := pio
	params[3] := iocmd
	okay := cog := COGNEW(@entry, @params) + 1

PUB stop
	if cog
		COGSTOP(cog - 1)
		cog := 0

PUB get_bitmap
	return @bitmap

DAT
		org	0
entry
		mov	ptr, PAR
		rdlong	bitmap_ptr, ptr
		add	ptr, #4
		rdlong	lcd_ptr, ptr
		add	ptr, #4
		rdlong	pio_ptr, ptr
		add	ptr, #4
		rdlong	cmd_ptr, ptr

command		rdlong	data, cmd_ptr
		mov	cmd, data
		shr	cmd, #24		WZ
	if_z	jmp	#command

		mov	port, data
		shr	port, #8
		and	port, #$ff
		cmp	port, #$f0		WZ, WC	' is any of the LCD ports touched?
	if_b	jmp	#command			' no, loop waiting
		cmp	cmd, #io_cmd_out	WZ	' output to port?
	if_z	jmp	#output
		cmp	cmd, #io_cmd_in		WZ	' input from port?
	if_z	jmp	#input
		jmp	#command			' don't care about the others
return_data
		andn	cmd, #$ff			' clear data byte
		or	cmd, data			' set input data
command_0
		and	cmd, low_24bits			' clear the command byte
		wrlong	cmd, cmd_ptr			' command finished (and return data)
		jmp	#command

'*******************************************
' Input from a LCD port
'
input
		mov	cmd, data
		cmp	port, #$ff		WZ	' input from data port?
	if_z	jmp	#lcd_data_r
		' fall through
'*******************************************
' LC display status read
' read row select?
'
lcd_status_r
		mov	data, #%01000000		' always return $40 ($80 would be "busy")
		jmp	#return_data

'*******************************************
' Output to a LCD port
'
output
		and	data, #$ff			' isolate data byte
		wrlong	data, cmd_ptr			' command finished
		cmp	port, #$ff		WZ	' output to data port?
	if_z	jmp	#lcd_data_w
		' fall through
'*******************************************
' LC display row select write
'
lcd_column_w
		rdlong	pio_port, pio_ptr		' get current PIO ports
		mov	t2, #1				' bit 0
		mov	t3, #10				' 10 columns to check
		mov	t4, #lcd_column			' offset of LCD columns in cog
:loop		test	pio_port, t2		WZ	' pio port A/B bit set?
	if_z	jmp	#:skip				' nope, skip
		movd	:column1, t4
		movd	:column2, t4
:column1	mov	0-0, data			' store lcd column
		mov	t1, data			' get lcd column
		and	t1, #%00111111			' mask bits 5..0
		cmp	t1, #LCD_DRIVER_W	WC	' >= 50?
:column2 if_nc	andn	0-0, #%00111111			' yes, clear bits 5..0
		or	lcd_status, t2			' set status bit = 1
:skip		add	t4, #1				' next column
		rol	t2, #1				' next bit
		djnz	t3, #:loop
		jmp	#command_0

'*******************************************
' LC display data read
'
lcd_data_r
		mov	data, #0			' assume no data
		rdlong	pio_port, pio_ptr		' get current PIO ports
		mov	t2, #1				' bit 0
		mov	t3, #10				' 10 columns to check
		mov	t4, #lcd_column			' offset of LCD columns in cog
:loop		test	pio_port, t2		WZ	' pio port A/B bit set?
	if_z	jmp	#:skip				' nope, skip
		movs	:column1, t4
		movd	:column2, t4
		movd	:column3, t4
		mov	t1, #10
		sub	t1, t3				' t1 = column 0..9
		shl	t1, #8				' * 256
:column1	add	t1, 0-0				' add lcd column
		add	t1, lcd_ptr			' add offset to LCD data in hub RAM
		rdbyte	data, t1			' read data byte
		sub	t1, lcd_ptr			' restore t1 = column
		test	lcd_status, t2		WZ	' status bit clear?
	if_z	add	t1, #1				' yes, column++
		and	t1, #$ff			' mask bits 7..0
:column2	mov	0-0, t1				' store (new) LCD column
		and	t1, #%00111111			' mask bits 5..0
		cmp	t1, #LCD_DRIVER_W	WC	' >= 50?
:column3 if_nc	andn	0-0, #%00111111			' yes, clear bits 5..0
		andn	lcd_status, t2			' clear status bit = 0
		jmp	#return_data			' exit on first read
:skip		add	t4, #1				' next column
		rol	t2, #1				' next bit
		djnz	t3, #:loop
		jmp	#return_data

'*******************************************
' LC display data write
'
lcd_data_w
		rdlong	pio_port, pio_ptr		' get current PIO ports
		mov	t2, #1				' bit 0
		mov	t3, #10				' 10 columns to check
		mov	t4, #lcd_column			' offset of LCD columns in cog
:loop		test	pio_port, t2		WZ	' pio port A/B bit set?
	if_z	jmp	#:skip				' nope, skip
		movs	:column1, t4
		movd	:column2, t4
		movd	:column3, t4
		mov	t1, #10
		sub	t1, t3				' t1 = column 0..9
		shl	t1, #8				' * 256
:column1	add	t1, 0-0				' add lcd column
		add	t1, lcd_ptr			' add offset to LCD data in hub RAM
		rdbyte	char, t1
		wrbyte	data, t1			' write data byte
		sub	t1, lcd_ptr			' restore t1 = column
		xor	char, data		WZ
	if_nz	call	#bitmap_draw
		add	t1, #1				' column++
		and	t1, #$ff			' mask bits 7..0
:column2	mov	0-0, t1				' store new LCD column
		and	t1, #%00111111			' mask bits 5..0
		cmp	t1, #LCD_DRIVER_W	WC	' >= 50?
:column3 if_nc	andn	0-0, #%00111111			' yes, clear bits 5..0
		andn	lcd_status, t2			' clear status bit = 0
:skip		add	t4, #1				' next column
		rol	t2, #1				' next bit
		djnz	t3, #:loop
		jmp	#command_0

bitmap_draw
		mov	ptr, #10
		sub	ptr, t3				' driver 0..9
		cmp	ptr, #5		WZ, WC
	if_ae	sub	ptr, #5				' 0..4
		shl	ptr, #3				' 0,8,16,24,32
	if_ae	add	ptr, s40x32			' add offset to lower half of display
		add	ptr, bitmap_ptr			' + bitmap base
		test	t1, #%00100000		WZ	' columns 0..31 or
	if_nz	add	ptr, #4				' columns 32..63?
		test	t1, #%01000000		WZ	' rows 1 or 3?
	if_nz	add	ptr, s40x8
		test	t1, #%10000000		WZ	' rows 2 or 3?
	if_nz	add	ptr, s40x16
		mov	bit, #1
		rol	bit, t1				' bit 0..31 depending on column
		mov	dbit, #1			' data bit mask
:loop		rdlong	bmap, ptr
		test	data, dbit		WZ
		muxnz	bmap, bit
		wrlong	bmap, ptr
		add	ptr, #40			' next scan row
		rol	dbit, #1			' next data bit
		cmp	dbit, #$100		WZ
	if_nz	jmp	#:loop
bitmap_draw_ret
		ret

'--------------------------------------------------------------------------------------------------
'constants > $1FF
low_24bits	long	$00ffffff
s40x32		long	40*32
s40x16		long	40*16
s40x8		long	40*8

bitmap_ptr	long	0-0
lcd_ptr		long	0-0
pio_ptr		long	0-0
cmd_ptr		long	0-0

lcd_column	long	0[10]				' current LCD column for the 10 drivers
lcd_status	long	0				' LCD status for the 10 drivers

t1		res	1
t2		res	1
t3		res	1
t4		res	1

pio_port	res	1				' current status of the 3 PIO ports
cmd		res	1				' i/o command
port		res	1				' i/o port
data		res	1				' i/o data
char		res	1
ptr		res	1
bit		res	1
dbit		res	1
bmap		res	1

		fit	$1f0
lcd		byte	0[10*256]			' 10 drivers
bitmap		byte	0[10*256]			' rotated and reassembled into scanlines

''      MIT LICENSE
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}
